#
#   based on meshOptMWE.cpp from Ketan Mittal
#
# python test_tmop.py -m ../data/square01.mesh -o 2 -rs 2 -mid 80 -tid 5 -ni 50 -qo 4 -vl 2 -ae 0
import sys
import os
from os.path import expanduser, join
import numpy as np

if len(sys.argv) > 1 and sys.argv[1] == '-p':
    import mfem.par as mfem
    use_parallel = True
    from mfem.common.mpi_debug import nicePrint as print
    from mpi4py import MPI
    myid = MPI.COMM_WORLD.rank
    sys.argv = [sys.argv[0]] + sys.argv[2:]

else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0

class discrete_size_2d(mfem.PyCoefficient):
    def EvalValue(self, x):
        opt = 2;
        small = 0.001
        big = 0.01
        val = 0.;
        
        xc = x[0] - 0.0
        yc = x[1] - 0.5
        r = np.sqrt(xc*xc + yc*yc)
        r1 = 0.45
        r2 = 0.55
        sf=30.0
        val = 0.5*(1+np.tanh(sf*(r-r1))) - 0.5*(1+np.tanh(sf*(r-r2)))

        val = max(0.,val)
        val = min(1.,val)

        return val * small + (1.0 - val) * big
    
def run(args):
    mesh_file = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))

    mesh_poly_deg = args.order
    rs_levels = args.refine_serial
    metric_id = args.metric_id
    target_id = args.target_id
    quad_type = args.quad_type
    quad_order = args.quad_order
    solver_type = args.solver_type
    lin_solver = args.lin_solver
    normalization = args.normalization

    verbosity_level = args.verbosity_level
    
    n_hr_iter=args.n_hr_iter
    n_h_iter=args.n_h_iter
   
    solver_iter = args.newton_iters
    solver_rtol= args.newton_rel_tolerance
    solver_art_type = args.adaptive_rel_tol
    max_lin_iter = args.lin_iter
    hradaptivity = args.hr_adaptivity
    visualization = not args.no_visualization
    adapt_eval= args.adaptivity_evaluator

    devopt    = "cpu";

    # 2. Initialize and refine the starting mesh.
    mesh = mfem.Mesh(mesh_file, 1, 1, False)
    for i in range(rs_levels):
        mesh.UniformRefinement()
    dim = mesh.Dimension();

    fec = mfem.H1_FECollection(mesh_poly_deg, dim)
    fespace = mfem.FiniteElementSpace(mesh, fec, dim)

    mesh.SetNodalFESpace(fespace)

    b = mfem.Vector(0)

    x = mfem.GridFunction(fespace)
    mesh.SetNodalGridFunction(x)
    x.SetTrueVector()
    x.SetFromTrueVector()
                              
    # 9. Save the starting (prior to the optimization) mesh to a file. This
    #    output can be viewed later using GLVis: "glvis -m perturbed.mesh".
    mesh.Print("perturbed.mesh")

    # 10. Store the starting (prior to the optimization) positions.
    x0 = mfem.GridFunction(fespace)
    x0.Assign(x)

    metric = mfem.tmop.TMOP_Metric_080(0.5)

    ind_fec = mfem.H1_FECollection(mesh_poly_deg, dim)
    ind_fes = mfem.FiniteElementSpace(mesh, ind_fec)
    size = mfem.GridFunction(ind_fes)

    
    if target_id == 5: # Discrete size 2D or 3D
         target_t = mfem.tmop.TargetConstructor.IDEAL_SHAPE_GIVEN_SIZE
                              
         tc = mfem.tmop.DiscreteAdaptTC(target_t)
                              
         if adapt_eval == 0:
             tc.SetAdaptivityEvaluator(mfem.tmop.AdvectorCG())
         else:
             if "InterpolatorFP" in dir(mfem.tmop):
                 evaluator = mfem.tmop.InterpolatorFP()
                 tc.SetAdaptivityEvaluator(evaluator)
             else:
                 assert False, "MFEM is not built with GSLIB."
         if dim == 2:
             #size_coeff = mfem.FunctionCoefficient(discrete_size_2d)
             size_coeff = discrete_size_2d()
             size.ProjectCoefficient(size_coeff)
         else:
             assert False, "only dim == 2 supported for this MWE."
                              
         tc.SetSerialDiscreteTargetSize(size)
         target_c = tc;

    else:
        print("Unknown target_id: " + str(target_id))
        return

    if target_c is None:
       target_c = mfem.tmop.TargetConstructor(target_t);

    target_c.SetNodes(x0)
    
    tmop_integ = mfem.tmop.TMOP_Integrator(metric, target_c)
 
    # Setup the quadrature rules for the TMOP integrator.
    if quad_type == 1:
        irules = mfem.IntegrationRules(0, mfem.Quadrature1D.GaussLobatto)
    elif quad_type == 2:        
        irules = mfem.IntRules
    elif quad_type == 3:        
        irules = mfem.IntegrationRules(0, mfem.Quadrature1D.ClosedUniform)
    else:
        print( "Unknown quad_type: " + str(quad_type))
        return 3

    tmop_integ.SetIntegrationRules(irules, quad_order)

    if normalization:
        tmop_integ.EnableNormalization(x0)

    a = mfem.NonlinearForm(fespace)
    a.AddDomainIntegrator(tmop_integ)

    # For HR tests, the energy is normalized by the number of elements.
    init_energy = a.GetGridFunctionEnergy(x);
    
    # Visualize the starting mesh and metric values.
    # Note that for combinations of metrics, this only shows the first metric.
    if visualization:
        mfem.tmop.vis_tmop_metric_s(mesh_poly_deg, metric, target_c, mesh, "Initial metric values", 0);

    # 13. Fix all boundary nodes, or fix only a given component depending on the
    #     boundary attributes of the given mesh. Attributes 1/2/3 correspond to
    #     fixed x/y/z components of the node. Attribute 4 corresponds to an
    #     entirely fixed node. Other boundary attributes do not affect the node
    #     movement boundary conditions.
    ess_bdr = mfem.intArray([1]*mesh.bdr_attributes.Max())
    a.SetEssentialBC(ess_bdr);


    # 14. As we use the Newton method to solve the resulting nonlinear system,
    #     here we setup the linear solver for the system's Jacobian.
    linsol_rtol = 1e-12;
    if lin_solver == 0:
        S = mfem.DSmoother(1, 1.0, max_lin_iter)
    elif lin_solver == 1:
        cg = mfem.CGSolver()
        cg.SetMaxIter(max_lin_iter)
        cg.SetRelTol(linsol_rtol)
        cg.SetAbsTol(0.0)
        cg.SetPrintLevel(3 if verbosity_level >= 2 else -1)
        S = cg
    else:
        minres = mfem.MINRESSolver()
        minres.SetMaxIter(max_lin_iter)
        minres.SetRelTol(linsol_rtol)
        minres.SetAbsTol(0.0)
        if verbosity_level > 2:
            minres.SetPrintLevel(1)
        minres.SetPrintLevel(3 if verbosity_level == 2 else -1)
        if lin_solver == 3 or lin_solver == 4:
             ds = mfem.DSmoother((0 if lin_solver == 3 else 1), 1.0, 1)
             ds.SetPositiveDiagonal(True)
             minres.SetPreconditioner(ds)
        S = minres;
   
    #/ Perform the nonlinear optimization.
    ir = irules.Get(fespace.GetFE(0).GetGeomType(), quad_order)
    solver = mfem.tmop.TMOPNewtonSolver(ir, solver_type)
    solver.SetIntegrationRules(irules, quad_order)
    if solver_type == 0:
        # Specify linear solver when we use a Newton-based solver.
        solver.SetPreconditioner(S)

    print(dir(solver))
    solver.SetMaxIter(solver_iter)
    solver.SetRelTol(solver_rtol)
    solver.SetAbsTol(0.0)
    if solver_art_type > 0:
        solver.SetAdaptiveLinRtol(solver_art_type, 0.5, 0.9)

    solver.SetPrintLevel(1 if verbosity_level >= 1 else -1)

    hr_solver = mfem.tmop.TMOPHRSolver(mesh, a, solver, x, False, hradaptivity,
                                  mesh_poly_deg, metric_id, n_hr_iter, n_h_iter)
    hr_solver.AddGridFunctionForUpdate(x0)
    hr_solver.Mult()

    # 15. Save the optimized mesh to a file. This output can be viewed later
    #     using GLVis: "glvis -m optimized.mesh".
    mesh.Print("optimized.mesh", 14)

    fin_energy = a.GetGridFunctionEnergy(x)
    print("Initial strain energy: " + "{:g}".format(init_energy))
    print("  Final strain energy: " + "{:g}".format(fin_energy))
    print("The strain energy decreased by: " + 
         "{:g}".format((init_energy - fin_energy) * 100.0 / init_energy))

    # 16. Visualize the final mesh and metric values.
    if visualization:
       mfem.tmop.vis_tmop_metric_s(mesh_poly_deg, metric, target_c, mesh, "Final metric values", 600);

    # 17. Visualize the mesh displacement.
    if visualization:
        x0 -= x
        sock = mfem.socketstream("localhost", 19916)
        sock << "solution\n" << mesh << x0
        sock.flush()
        sock << "window_title 'Displacements'\n" << "window_geometry "
        sock << 1200 << " " << 0 << " " << 600 << " " << 600 << "\n"
        sock << "keys jRmclA"
        sock.flush()             


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='meshOptMWE')
    parser.add_argument('-m', '--mesh',
                        default='square01.mesh',  # icf.mesh
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree) or -1 for isoparametric space.")
    parser.add_argument('-rs', '--refine-serial',
                        action='store', default=2, type=int,
                        help="Number of times to refine the mesh uniformly in serial")
    parser.add_argument("-mid", "--metric-id",
                        action="store", default=80, type=int,
                        help="\n".join(["Mesh optimization metric:",
                                        "\tT-metrics",
                                        "2  : 0.5|T|^2/tau-1                 -- 2D shape (condition number)\n\t"]))
    parser.add_argument("-tid", "--target-id",
                        action="store", default=5, type=int,                        
                        help="\n".join(["Target (ideal element) type:",
                                        "\t5: Ideal shape, given size (in physical space)"]))
    parser.add_argument("-qt", "--quad-type",
                        action="store", default=1, type=int,                                                
                        help="\n".join(["Quadrature rule type:",
                                        "\t1: Gauss-Lobatto",
                                        "\t2: Gauss-Legendre"
                                        "\t3: Closed uniform points"]))
    parser.add_argument("-qo", "--quad_order",
                        action="store", default=4, type=int,                                                
                         help="Order of the quadrature rule.")
    parser.add_argument("-st", "--solver-type",
                        action="store", default=0, type=int,
                        help = " Type of solver: (default) 0: Newton, 1: LBFGS")
    parser.add_argument("-ni", "--newton-iters",
                        action="store", default=80, type=int,                        
                        help="Maximum number of Newton iterations.")
    parser.add_argument("-rtol", "--newton-rel-tolerance",
                        action="store", default=1e-10, type=float,
                        help="Relative tolerance for the Newton solver.")
    parser.add_argument("-art", "--adaptive-rel-tol",
                        action="store", default=0, type=int,                        
                        help="\n".join(["Type of adaptive relative linear solver tolerance:",
                                        "\t0: None (default)",
                                        "\t1: Eisenstat-Walker type 1",
                                        "\t2: Eisenstat-Walker type 2"]))
    parser.add_argument("-ls", "--lin-solver",
                        action="store", default=2, type=int,                                                 
                        help="\n".join(["Linear solver:",
                                        "\t0: l1-Jacobi",
                                        "\t1: CG",
                                        "\t2: MINRES",
                                        "\t3: MINRES + Jacobi preconditioner",
                                        "\t4: MINRES + l1-Jacobi preconditioner"]))
    parser.add_argument("-li", "--lin-iter",
                        action="store", default=100, type=int,     
                        help="Maximum number of iterations in the linear solve.")
    parser.add_argument("-hr", "--hr-adaptivity", 
                        action='store_true',
                        help="Enable hr-adaptivity.")
    parser.add_argument("-nor", "--normalization", 
                        action='store_true',
                        help="Make all terms in the optimization functional unitless.")
    parser.add_argument('-no-vis', '--no-visualization',
                        action='store_true',
                        help='Enable GLVis visualization')
    parser.add_argument("-vl", "--verbosity-level",
                        action="store", default=2, type=int,                                                 
                        help="Set the verbosity level - 0, 1, or 2.")


    parser.add_argument("-ae", "--adaptivity-evaluator",
                        action="store", default=1, type=int,                                                 
                        help="0 - Advection based (DEFAULT), 1 - GSLIB.");

    parser.add_argument("-nhr", "--n_hr_iter",
                        action="store", default=5, type=int,
                        help="Number of hr-adaptivity iterations.")

    parser.add_argument("-nh", "--n_h_iter",
                        action="store", default=1, type=int,
                        help="Number of h-adaptivity iterations per r-adaptivity")
                  
    args = parser.parse_args()
    parser.print_options(args)
                  
    run(args)



try:
   import mfem.par as mfem
   MFEM_PAR = True
except:
   import mfem.ser as mfem    
   MFEM_PAR = False

from scipy.sparse import csr_matrix, coo_matrix, lil_matrix, csc_matrix

def eliminate_rows(m, rows):
    '''
    if m is not csr, it is converted to csr
    return csr_matrix
    '''
    if not isinstance(m, csr_matrix):
        try:
            m = m.tocsr()
        except:
            raise ValueError('Can not convert Matrix to CSR format.')            
    for row in rows:
        m.data[m.indptr[row]:m.indptr[row+1]] = 0.0  #  m[x,:] = 0.0
    m.eliminate_zeros()
    return m

def eliminate_cols(m, cols):
    '''
    if m is not csr, it is converted to csc
    return csc_matrix
    '''
    if not isinstance(m, csc_matrix):
        try:
            m = m.tocsc()
        except:
            raise ValueError('Can not convert Matrix to CSR format.')            
    for col in cols:
       m.data[m.indptr[col]:m.indptr[col+1]] = 0.0  #  m[:,x] = 0.0
    m.eliminate_zeros()
    return m

def sparsemat_to_scipycsr(mat, dtype):
     w, h = mat.Width(), mat.Height()
     I = mat.GetIArray()
     J = mat.GetJArray()
     data = mat.GetDataArray()
     mat.LoseData()
     m1 =csr_matrix((data, J, I), shape = (h, w),
                    dtype = dtype)
     return m1
        




from __future__ import print_function
import os
from os.path import expanduser, join
import sys
import numpy as np
import io
from mfem import path as mfem_path

if len(sys.argv) > 1 and sys.argv[1] == '-p':   
    import mfem.par as mfem
    use_parallel = True
    from mfem.common.mpi_debug import nicePrint as print
    from mpi4py import MPI
    myid  = MPI.COMM_WORLD.rank
    
else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0

def run_test():
    #meshfile = expanduser(join(mfem_path, 'data', 'semi_circle.mesh'))
    dir = os.path.dirname(os.path.abspath(__file__))
    meshfile = os.path.join(dir, "../data/amr-quad.mesh")
    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()
    sdim = mesh.SpaceDimension()    
    fec = mfem.H1_FECollection(1, dim)
    fespace = mfem.FiniteElementSpace(mesh, fec, 1)
    print('Number of finite element unknowns: '+
          str(fespace.GetTrueVSize()))
    
    c = mfem.ConstantCoefficient(1.0)
    gf = mfem.GridFunction(fespace)
    gf.ProjectCoefficient(c)

    print("write mesh to STDOUT")
    mesh.Print(mfem.STDOUT)
    print("creat VTK file to file")
    mesh.PrintVTK('mesh.vtk', 1)
    print("creat VTK to STDOUT")
    mesh.PrintVTK(mfem.STDOUT, 1)
    print("save GridFunction to file")
    gf.Save('out_test_gridfunc.gf')
    gf.SaveVTK(mfem.wFILE('out_test_gridfunc1.vtk'), 'data', 1)
    print("save GridFunction to file in VTK format")    
    gf.SaveVTK('out_test_gridfunc2.vtk', 'data', 1)
    print("Gridfunction to STDOUT")
    gf.Save(mfem.STDOUT)
    
    o = io.StringIO()
    count = gf.Save(o)
    count2 = gf.SaveVTK(o, 'data', 1)
    print("length of data ", count, count2)
    print('result: ', o.getvalue())    
    
if __name__=='__main__':
    run_test()



from __future__ import print_function
import os
import sys
import numpy as np

if len(sys.argv) > 1 and sys.argv[1] == '-p':
   print("loading parallel")
   import mfem.par as mfem
else:
   import mfem.ser as mfem
        

def run_test():
    print("Test sparsemat module")
    indptr = np.array([0, 2, 3, 6], dtype=np.int32)
    indices = np.array([0, 2, 2, 0, 1, 2], dtype=np.int32)
    data = np.array([1, 2, 3, 4, 5, 6], dtype=float)
    
    mat = mfem.SparseMatrix([indptr, indices, data, 3,3])
    mat.Print()
    mat.PrintInfo()
    
if __name__=='__main__':
    run_test()



'''
   MFEM example 12p

   See c++ version in the MFEM library for more detail 

   How to run:
      mpirun -np 2 python <arguments>
  
   Example of arguments:
      ex12p.py -m beam-tri.mesh
      ex12p.py -m beam-quad.mesh
      ex12p.py -m beam-tet.mesh -n 10 -o 2 -elast
      ex12p.py -m beam-hex.mesh
      ex12p.py -m beam-tri.mesh -o 2 -sys
      ex12p.py -m beam-quad.mesh -n 6 -o 3 -elast
      ex12p.py -m beam-quad-nurbs.mesh
      ex12p.py -m beam-hex-nurbs.mesh
'''
import sys
from os.path import expanduser, join, dirname
import numpy as np

from mfem.common.arg_parser import ArgParser
import mfem.par as mfem
from mpi4py import MPI

num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '{:0>6d}'.format(myid)

parser = ArgParser(description='Ex12 (linear elastisity)')
parser.add_argument('-m', '--mesh',
                    default='beam-tri.mesh',
                    action='store', type=str,
                    help='Mesh file to use.')
parser.add_argument('-o', '--order',
                    action='store', default=1, type=int,
                    help="Finite element order (polynomial degree)")
parser.add_argument("-n", "--num-eigs",
                    action='store', default=5, type=int,
                    help="Number of desired eigenmodes.")
parser.add_argument("-s", "--seed",
                    action='store', default=66, type=int,
                    help="Random seed used to initialize LOBPCG.")
parser.add_argument('-elast', '--amg-for-elasticity',
                    action='store_true',
                    help='Use the special AMG elasticity solver (GM/LN approaches)',
                    dest='amg_elast', default=False)
parser.add_argument('-sys', '--amg-for-systems',
                    action='store_false',
                    help='Use  standard AMG for systems (unknown approach).',
                    dest='amg_elast', default=True)
parser.add_argument('-vis', '--visualization',
                    action='store_true', default=True,
                    help='Enable GLVis visualization')
args = parser.parse_args()

order = args.order
amg_elast = args.amg_elast
nev = args.num_eigs
visualization = args.visualization
seed = args.seed
if (myid == 0):
    parser.print_options(args)

device = mfem.Device('cpu')
if myid == 0:
    device.Print()

# 3. Read the mesh from the given mesh file on all processors. We can handle
#    triangular, quadrilateral, tetrahedral, hexahedral, surface and volume
#    meshes with the same code
meshfile = expanduser(join(dirname(__file__), '..', 'data', args.mesh))
mesh = mfem.Mesh(meshfile, 1, 1)
dim = mesh.Dimension()
if (mesh.attributes.Max() < 2):
    if (myid == 0):
        print("Input mesh should have at least two materials!")
        print(" (See schematic in ex12p.cpp)")
    sys.exit()

# 4. Select the order of the finite element discretization space. For NURBS
#    meshes, we increase the order by degree elevation.
if (mesh.NURBSext and order > mesh.NURBSext.GetOrder()):
    mesh.DegreeElevate(order - mesh.NURBSext.GetOrder())
#  5. Refine the serial mesh on all processors to increase the resolution. In
#     this example we do 'ref_levels' of uniform refinement. We choose
#     'ref_levels' to be the largest number that gives a final mesh with no
#     more than 1,000 elements.
ref_levels = int(np.floor(np.log(1000./mesh.GetNE())/np.log(2.)/dim))
for x in range(ref_levels):
    mesh.UniformRefinement()

#  6. Define a parallel mesh by a partitioning of the serial mesh. Refine
#     this mesh further in parallel to increase the resolution. Once the
#     parallel mesh is defined, the serial mesh can be deleted.
pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
del mesh
par_ref_levels = 1
for l in range(par_ref_levels):
    pmesh.UniformRefinement()

# 7. Define a parallel finite element space on the parallel mesh. Here we
#    use vector finite elements, i.e. dim copies of a scalar finite element
#    space. We use the ordering by vector dimension (the last argument of
#    the FiniteElementSpace constructor) which is expected in the systems
#    version of BoomerAMG preconditioner. For NURBS meshes, we use the
#    (degree elevated) NURBS space associated with the mesh nodes.
use_nodal_fespace = pmesh.NURBSext and not amg_elast
if use_nodal_fespace:
    fespace = pmesh.GetNodes().FESpace()
else:
    fec = mfem.H1_FECollection(order, dim)
    fespace = mfem.ParFiniteElementSpace(pmesh, fec, dim,
                                         mfem.Ordering.byVDIM)
size = fespace.GlobalTrueVSize()
if (myid == 0):
    print("Number of unknowns: " + str(size))
    print("Assembling")

# 8. Set up the parallel bilinear forms a(.,.) and m(.,.) on the finite
#    element space corresponding to the linear elasticity integrator with
#    piece-wise constants coefficient lambda and mu, a simple mass matrix
#    needed on the right hand side of the generalized eigenvalue problem
#    below. The boundary conditions are implemented by marking only boundary
#    attribute 1 as essential. We use special values on the diagonal to
#    shift the Dirichlet eigenvalues out of the computational range. After
#    serial/parallel assembly we extract the corresponding parallel matrices
#    A and M.
lamb = mfem.Vector(pmesh.attributes.Max())
lamb.Assign(1.0)
lamb[0] = lamb[1]*50
lambda_func = mfem.PWConstCoefficient(lamb)
mu = mfem.Vector(pmesh.attributes.Max())
mu.Assign(1.0)
mu[0] = mu[1]*50
mu_func = mfem.PWConstCoefficient(mu)

ess_bdr = mfem.intArray(pmesh.bdr_attributes.Max())
ess_bdr.Assign(0)
ess_bdr[0] = 1

a = mfem.ParBilinearForm(fespace)
a.AddDomainIntegrator(mfem.ElasticityIntegrator(lambda_func, mu_func))
if (myid == 0):
    print("matrix ... ")
a.Assemble()
a.EliminateEssentialBCDiag(ess_bdr, 1.0)
a.Finalize()

m = mfem.ParBilinearForm(fespace)
m.AddDomainIntegrator(mfem.VectorMassIntegrator())
m.Assemble()

# shift the eigenvalue corresponding to eliminated dofs to a large value
m.EliminateEssentialBCDiag(ess_bdr,  sys.float_info.min)
m.Finalize()
if (myid == 0):
    print("done. ")

A = a.ParallelAssemble()
M = m.ParallelAssemble()

# 9. Define and configure the LOBPCG eigensolver and the BoomerAMG
#    preconditioner for A to be used within the solver. Set the matrices
#    which define the generalized eigenproblem A x = lambda M x.
amg = mfem.HypreBoomerAMG(A)
amg.SetPrintLevel(0)
if (amg_elast):
    amg.SetElasticityOptions(fespace)
else:
    amg.SetSystemsOptions(dim)

lobpcg = mfem.HypreLOBPCG(MPI.COMM_WORLD)
lobpcg.SetNumModes(nev)
lobpcg.SetRandomSeed(seed)
lobpcg.SetPreconditioner(amg)
lobpcg.SetMaxIter(100)
lobpcg.SetTol(1e-8)
lobpcg.SetPrecondUsageMode(1)
lobpcg.SetPrintLevel(1)
lobpcg.SetMassMatrix(M)
lobpcg.SetOperator(A)

# 10. Compute the eigenmodes and extract the array of eigenvalues. Define a
#     parallel grid function to represent each of the eigenmodes returned by
#     the solver.
eigenvalues = mfem.doubleArray()
lobpcg.Solve()
lobpcg.GetEigenvalues(eigenvalues)
x = mfem.ParGridFunction(fespace)


# 11. For non-NURBS meshes, make the mesh curved based on the finite element
#     space. This means that we define the mesh elements through a fespace
#     based transformation of the reference element. This allows us to save
#     the displaced mesh as a curved mesh when using high-order finite
#     element displacement field. We assume that the initial mesh (read from
#     the file) is not higher order curved mesh compared to the chosen FE
#     space.
if not use_nodal_fespace:
    pmesh.SetNodalFESpace(fespace)

# 12. Save the refined mesh and the modes in parallel. This output can be
#     viewed later using GLVis: "glvis -np <np> -m mesh -g mode".
smyid = '{:0>6d}'.format(myid)
mesh_name = "mesh."+smyid
pmesh.Print(mesh_name, 8)

for i in range(nev):
    x.Assign(lobpcg.GetEigenvector(i))
    sol_name = "mode_"+str(i).zfill(2)+"."+smyid
    x.Save(sol_name, 8)

# 13. Send the above data by socket to a GLVis server. Use the "n" and "b"
#     keys in GLVis to visualize the displacements.
if (visualization):
    mode_sock = mfem.socketstream("localhost", 19916)
    mode_sock.precision(8)

    for i in range(nev):
        if (myid == 0):
            print("Eigenmode " + str(i+1) + '/' + str(nev) +
                  ", Lambda = " + str(eigenvalues[i]))

        # convert eigenvector from HypreParVector to ParGridFunction
        x.Assign(lobpcg.GetEigenvector(i))

        mode_sock.send_text("parallel " + str(num_procs) + " " + str(myid))
        mode_sock.send_solution(pmesh,   x)
        mode_sock.send_text("window_title 'Eigenmode " + str(i+1) + '/' +
                            str(nev) + ", Lambda = " + str(eigenvalues[i]) + "'")

        c = None
        if (myid == 0):
            from builtins import input
            c = input("press (q)uit or (c)ontinue --> ")
        c = MPI.COMM_WORLD.bcast(c, root=0)
        if (c != 'c'):
            break

    mode_sock.close()



'''
   MFEM example 39 (converted from ex39.cpp)

   See c++ version in the MFEM library for more detail

   Sample runs: python ex39.py
                python ex39.py -ess "Southern Boundary"
                python ex39.py -src Base

   Description:  This example code demonstrates the use of named attribute
                 sets in MFEM to specify material regions, boundary regions,
                 or source regions by name rather than attribute numbers. It
                 also demonstrates how new named attribute sets may be created
                 from arbitrary groupings of attribute numbers and used as a
                 convenient shorthand to refer to those groupings in other
                 portions of the application or through the command line.
  
                 The particular problem being solved here is nearly the same
                 as that in example 1 i.e. a simple finite element
                 discretization of the Laplace problem -Delta u = 1 with
                 homogeneous Dirichlet boundary conditions and, in this case,
                 an inhomogeneous diffusion coefficient. The diffusion
                 coefficient is given a small default value throughout the
                 domain which is increased by two separate amounts in two named
                 regions.
  
                 This example makes use of a specific input mesh, "compass.msh",
                 containing named domain and boundary regions generated by Gmsh
                 and stored in their "msh" format (version 2.2). This file
                 defines eight boundary regions corresponding to eight compass
                 headings; "ENE", "NNE", "NNW", "WSW", "SSW", "SSE", and "ESE".
                 It also defines nine domain regions; "Base", "N Even", "N Odd",
                 "W Even", "W Odd", "S Even", "S Odd", "E Even", and "E Odd".
                 These regions split the four compass pointers into two halves
                 each and also label the remaining elements as "Base". Starting
                 with these named regions we test the construction of named
                 sets as well as reading and writing these named groupings from
                 and to mesh files.
  
                 The example highlights the use of named attribute sets for
                 both subdomains and boundaries in different contexts as well
                 as basic methods to create named sets from existing attributes.

'''
import os
from os.path import expanduser, join
import numpy as np

import mfem.ser as mfem


def run(order=1,
        meshfile='',
        source_name='Rose Even',
        ess_name='Boundary',
        visualization=True):

    # 2. Read the mesh from the given mesh file. We can handle triangular,
    #    quadrilateral, tetrahedral, hexahedral, surface and volume meshes with
    #    the same code.

    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()

    # 3. Refine the mesh to increase the resolution. In this example we do
    #    'ref_levels' of uniform refinement. We choose 'ref_levels' to be the
    #    largest number that gives a final mesh with no more than 50,000
    #    elements.
    ref_levels = int(np.log(50000./mesh.GetNE())/np.log(2.)/dim)
    for i in range(ref_levels):
        mesh.UniformRefinement()

    # 4a. Display attribute set names contained in the initial mesh
    #     GetAttributeSetNames returns Python set object.
    attr_sets = mesh.attribute_sets
    bdr_attr_sets = mesh.bdr_attribute_sets
    names = attr_sets.GetAttributeSetNames()
    print("Element Attribute Set Names: " +
          ' '.join(['"' + x + '"' for x in names]))
    names = bdr_attr_sets.GetAttributeSetNames()
    print("Boundary Attribute Set Names: " +
          ' '.join(['"' + x + '"' for x in names]))

    # 4b. Define new regions based on existing attribute sets
    Na = attr_sets.GetAttributeSet("N Even")
    Nb = attr_sets.GetAttributeSet("N Odd")
    Sa = attr_sets.GetAttributeSet("S Even")
    Sb = attr_sets.GetAttributeSet("S Odd")
    Ea = attr_sets.GetAttributeSet("E Even")
    Eb = attr_sets.GetAttributeSet("E Odd")
    Wa = attr_sets.GetAttributeSet("W Even")
    Wb = attr_sets.GetAttributeSet("W Odd")

    # Create a new set spanning the North point
    attr_sets.SetAttributeSet("North", Na)
    attr_sets.AddToAttributeSet("North", Nb)

    # Create a new set spanning the South point
    attr_sets.SetAttributeSet("South", Sa)
    attr_sets.AddToAttributeSet("South", Sb)

    # Create a new set spanning the East point
    attr_sets.SetAttributeSet("East", Ea)
    attr_sets.AddToAttributeSet("East", Eb)

    # Create a new set spanning the West point
    attr_sets.SetAttributeSet("West", Wa)
    attr_sets.AddToAttributeSet("West", Wb)

    # Create a new set consisting of the "a" sides of the compass rose
    attr_sets.SetAttributeSet("Rose Even", Na)
    attr_sets.AddToAttributeSet("Rose Even", Sa)
    attr_sets.AddToAttributeSet("Rose Even", Ea)
    attr_sets.AddToAttributeSet("Rose Even", Wa)

    # Create a new set consisting of the "b" sides of the compass rose
    attr_sets.SetAttributeSet("Rose Odd", Nb)
    attr_sets.AddToAttributeSet("Rose Odd", Sb)
    attr_sets.AddToAttributeSet("Rose Odd", Eb)
    attr_sets.AddToAttributeSet("Rose Odd", Wb)

    # Create a new set consisting of the full compass rose
    Ra = attr_sets.GetAttributeSet("Rose Even")
    Rb = attr_sets.GetAttributeSet("Rose Odd")
    attr_sets.SetAttributeSet("Rose", Ra)
    attr_sets.AddToAttributeSet("Rose", Rb)

    # 4c. Define new boundary regions based on existing boundary attribute sets
    NNE = bdr_attr_sets.GetAttributeSet("NNE")
    NNW = bdr_attr_sets.GetAttributeSet("NNW")
    ENE = bdr_attr_sets.GetAttributeSet("ENE")
    ESE = bdr_attr_sets.GetAttributeSet("ESE")
    SSE = bdr_attr_sets.GetAttributeSet("SSE")
    SSW = bdr_attr_sets.GetAttributeSet("SSW")
    WNW = bdr_attr_sets.GetAttributeSet("WNW")
    WSW = bdr_attr_sets.GetAttributeSet("WSW")

    bdr_attr_sets.SetAttributeSet("Northern Boundary", NNE)
    bdr_attr_sets.AddToAttributeSet("Northern Boundary", NNW)

    bdr_attr_sets.SetAttributeSet("Southern Boundary", SSE)
    bdr_attr_sets.AddToAttributeSet("Southern Boundary", SSW)

    bdr_attr_sets.SetAttributeSet("Eastern Boundary", ENE)
    bdr_attr_sets.AddToAttributeSet("Eastern Boundary", ESE)

    bdr_attr_sets.SetAttributeSet("Western Boundary", WNW)
    bdr_attr_sets.AddToAttributeSet("Western Boundary", WSW)

    bdr_attr_sets.SetAttributeSet("Boundary",
                                  bdr_attr_sets.GetAttributeSet
                                  ("Northern Boundary"))
    bdr_attr_sets.AddToAttributeSet("Boundary",
                                    bdr_attr_sets.GetAttributeSet
                                    ("Southern Boundary"))
    bdr_attr_sets.AddToAttributeSet("Boundary",
                                    bdr_attr_sets.GetAttributeSet
                                    ("Eastern Boundary"))
    bdr_attr_sets.AddToAttributeSet("Boundary",
                                    bdr_attr_sets.GetAttributeSet
                                    ("Western Boundary"))

    # 5. Define a finite element space on the mesh. Here we use continuous
    #    Lagrange finite elements of the specified order.

    fec = mfem.H1_FECollection(order, dim)
    fespace = mfem.FiniteElementSpace(mesh, fec)
    print('Number of finite element unknowns: ' +
          str(fespace.GetTrueVSize()))

    # 6. Determine the list of true (i.e. conforming) essential boundary dofs.
    #    In this example, the boundary conditions are defined by marking all
    #    the boundary regions corresponding to the boundary attributes
    #    contained in the set named "ess_name" as essential (Dirichlet) and
    #    converting them to a list of true dofs.
    ess_tdof_list = mfem.intArray()
    if bdr_attr_sets.AttributeSetExists(ess_name):
        ess_bdr_marker = bdr_attr_sets.GetAttributeSetMarker(ess_name)
        fespace.GetEssentialTrueDofs(ess_bdr_marker, ess_tdof_list)

    # 7. Set up the linear form b(.) which corresponds to the right-hand side of
    #    the FEM linear system, which in this case is (1_s,phi_i) where phi_i
    #    are the basis functions in fespace and 1_s is an indicator function
    #    equal to 1 on the region defined by the named set "source_name" and
    #    zero elsewhere.

    source_marker = attr_sets.GetAttributeSetMarker(source_name)
    b = mfem.LinearForm(fespace)
    one = mfem.ConstantCoefficient(1.0)
    b.AddDomainIntegrator(mfem.DomainLFIntegrator(one), source_marker)
    b.Assemble()

    # 8. Define the solution vector x as a finite element grid function
    #    corresponding to fespace. Initialize x with initial guess of zero,
    #    which satisfies the boundary conditions.
    x = mfem.GridFunction(fespace)
    x.Assign(0.0)

    # 9. Set up the bilinear form a(.,.) on the finite element space
    #    corresponding to the Laplacian operator -Delta, by adding the
    #    Diffusion domain integrator.
    a = mfem.BilinearForm(fespace)

    defaultCoef = mfem.ConstantCoefficient(1.0e-6)
    baseCoef = mfem.ConstantCoefficient(1.0)
    roseCoef = mfem.ConstantCoefficient(2.0)

    base_marker = attr_sets.GetAttributeSetMarker("Base")
    rose_marker = attr_sets.GetAttributeSetMarker("Rose Even")

    # Impose a very small diffusion coefficient across the entire mesh
    a.AddDomainIntegrator(mfem.DiffusionIntegrator(defaultCoef))

    # Impose an additional, stronger diffusion coefficient in select regions
    a.AddDomainIntegrator(mfem.DiffusionIntegrator(baseCoef), base_marker)
    a.AddDomainIntegrator(mfem.DiffusionIntegrator(roseCoef), rose_marker)

    # 10. Assemble the bilinear form and the corresponding linear system,
    #     applying any necessary transformations.
    a.Assemble()

    A = mfem.SparseMatrix()
    B = mfem.Vector()
    X = mfem.Vector()
    a.FormLinearSystem(ess_tdof_list, x, b, A, X, B)
    print("Size of linear system: " + str(A.Height()))

    # 11. Solve the system using PCG with symmetric Gauss-Seidel preconditioner.
    M = mfem.GSSmoother(A)
    mfem.PCG(A, M, B, X, 1, 800, 1e-12, 0.0)

    # 12. Recover the solution as a finite element grid function.
    a.RecoverFEMSolution(X, b, x)

    # 13. Save the refined mesh and the solution. This output can be viewed
    #     later using GLVis: "glvis -m refined.mesh -g sol.gf".
    mesh.Save('refined.mesh')
    x.Save('sol.gf')


    # 14. Send the solution by socket to a GLVis server.
    if visualization:
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock.precision(8)
        sol_sock << "solution\n" << mesh << x << "keys Rjmm"
        sol_sock.flush()


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex39 (Named attribute)')
    parser.add_argument('-m', '--mesh',
                        default='compass.msh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help='Finite element order (polynomial degree) or -1 for isoparametric space.')
    parser.add_argument('-src', '--source-attr-name',
                        action='store', default='Rose Even', type=str,
                        help='Name of attribute set containing source.')
    parser.add_argument('-ess', '--ess-attr-name',
                        action='store', default='Boundary', type=str,
                        help='Name of attribute set containing essential BC.')
    parser.add_argument('-no-vis', '--no-visualization',
                        action='store_true',
                        default=False,
                        help='Disable or disable GLVis visualization')

    args = parser.parse_args()
    parser.print_options(args)

    order = args.order
    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))

    visualization = not args.no_visualization
    source_name = args.source_attr_name
    ess_name = args.ess_attr_name

    run(order=order,
        meshfile=meshfile,
        source_name=source_name,
        ess_name=ess_name,
        visualization=visualization)



'''
   MFEM example 1p

   See c++ version in the MFEM library for more detail 
'''
import sys
import os
from os.path import expanduser, join
import numpy as np

from mpi4py import MPI

from mfem.common.arg_parser import ArgParser
from mfem import path
import mfem.par as mfem

def_meshfile = expanduser(join(path, 'data', 'star.mesh'))

num_proc = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank


def run(order=1, static_cond=False,
        meshfile=def_meshfile,
        visualization=False,
        device='cpu',
        pa=False):

    algebraic_ceed = False

    device = mfem.Device(device)
    if myid == 0:
        device.Print()

    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()

    ref_levels = int(np.floor(np.log(10000./mesh.GetNE())/np.log(2.)/dim))
    for x in range(ref_levels):
        mesh.UniformRefinement()
    mesh.ReorientTetMesh()
    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
    del mesh

    par_ref_levels = 2
    for l in range(par_ref_levels):
        pmesh.UniformRefinement()

    if order > 0:
        fec = mfem.H1_FECollection(order, dim)
    elif mesh.GetNodes():
        fec = mesh.GetNodes().OwnFEC()
        print("Using isoparametric FEs: " + str(fec.Name()))
    else:
        order = 1
        fec = mfem.H1_FECollection(order, dim)

    fespace = mfem.ParFiniteElementSpace(pmesh, fec)
    fe_size = fespace.GlobalTrueVSize()

    if (myid == 0):
        print('Number of finite element unknowns: ' + str(fe_size))

    ess_tdof_list = mfem.intArray()
    if pmesh.bdr_attributes.Size() > 0:
        ess_bdr = mfem.intArray(pmesh.bdr_attributes.Max())
        ess_bdr.Assign(1)
        fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

    #   the basis functions in the finite element fespace.
    b = mfem.ParLinearForm(fespace)
    one = mfem.ConstantCoefficient(1.0)
    b.AddDomainIntegrator(mfem.DomainLFIntegrator(one))
    b.Assemble()

    x = mfem.ParGridFunction(fespace)
    x.Assign(0.0)

    a = mfem.ParBilinearForm(fespace)
    if pa:
        a.SetAssemblyLevel(mfem.AssemblyLevel_PARTIAL)

    a.AddDomainIntegrator(mfem.DiffusionIntegrator(one))

    if static_cond:
        a.EnableStaticCondensation()

    a.Assemble()

    A = mfem.OperatorPtr()
    B = mfem.Vector()
    X = mfem.Vector()
    a.FormLinearSystem(ess_tdof_list, x, b, A, X, B)

    if pa:
        if mfem.UsesTensorBasis(fespace):
            if algebraic_ceed:
                assert False, "not supported"
                # prec = new ceed::AlgebraicSolver(a, ess_tdof_list);
            else:
                prec = mfem.OperatorJacobiSmoother(a, ess_tdof_list)
    else:
        prec = mfem.HypreBoomerAMG()

    cg = mfem.CGSolver(MPI.COMM_WORLD)
    cg.SetRelTol(1e-12)
    cg.SetMaxIter(200)
    cg.SetPrintLevel(1)
    cg.SetPreconditioner(prec)
    cg.SetOperator(A.Ptr())
    cg.Mult(B, X)

    a.RecoverFEMSolution(X, b, x)

    smyid = '{:0>6d}'.format(myid)
    mesh_name = "mesh."+smyid
    sol_name = "sol."+smyid

    pmesh.Print(mesh_name, 8)
    x.Save(sol_name, 8)


if __name__ == "__main__":
    parser = ArgParser(description='Ex1 (Laplace Problem)')
    parser.add_argument('-m', '--mesh',
                        default='star.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree) or -1 for isoparametric space.")
    parser.add_argument('-sc', '--static-condensation',
                        action='store_true',
                        help="Enable static condensation.")
    parser.add_argument("-pa", "--partial-assembly",
                        action='store_true',
                        help="Enable Partial Assembly.")
    parser.add_argument("-d", "--device",
                        default="cpu", type=str,
                        help="Device configuration string, see Device::Configure().")

    args = parser.parse_args()

    if myid == 0:
        parser.print_options(args)

    order = args.order
    static_cond = args.static_condensation
    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))
    visualization = args.visualization
    device = args.device
    pa = args.partial_assembly

    run(order=order,
        static_cond=static_cond,
        meshfile=meshfile,
        visualization=visualization,
        device=device,
        pa=pa)



'''
   MFEM example 8p 

   See c++ version in the MFEM library for more detail 

   How to run:
      mpirun -np 2 python <arguments>

   Example of arguments:
      ex8p.py -m square-disc.mesh
      ex8p.py -m star.mesh
      ex8p.py -m escher.mesh
      ex8p.py -m fichera.mesh
      ex8p.py -m square-disc-p3.mesh
      ex8p.py -m star-surf.mesh -o 2
'''
import sys
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sin, cos, exp, sqrt

from mfem.common.arg_parser import ArgParser

import mfem.par as mfem
from mpi4py import MPI
num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank


parser = ArgParser(description='Ex8p')
parser.add_argument('-m', '--mesh',
                    default='star.mesh',
                    action='store', type=str,
                    help='Mesh file to use.')
parser.add_argument('-o', '--order',
                    action='store', default=1, type=int,
                    help="Finite element order (polynomial degree)")
parser.add_argument('-vis', '--visualization',
                    action='store_true', default=True,
                    help='Enable GLVis visualization')

args = parser.parse_args()

device = mfem.Device('cpu')
if myid == 0:
    device.Print()

order = args.order
visualization = args.visualization
if myid == 0:
    parser.print_options(args)

# 3. Read the (serial) mesh from the given mesh file on all processors.  We
#    can handle triangular, quadrilateral, tetrahedral, hexahedral, surface
#    and volume meshes with the same code.

meshfile = expanduser(join(dirname(__file__), '..', 'data', 'star.mesh'))
mesh = mfem.Mesh(meshfile, 1, 1)
dim = mesh.Dimension()

# 4. Refine the serial mesh on all processors to increase the resolution. In
#    this example we do 'ref_levels' of uniform refinement. We choose
#    'ref_levels' to be the largest number that gives a final mesh with no
#    more than 10,000 elements.
ref_levels = int(np.floor(np.log(10000./mesh.GetNE())/np.log(2.)/dim))
for x in range(ref_levels):
    mesh.UniformRefinement()

# 5. Define a parallel mesh by a partitioning of the serial mesh. Refine
#    this mesh further in parallel to increase the resolution. Once the
#    parallel mesh is defined, the serial mesh can be deleted.
pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
del mesh
par_ref_levels = 1
for l in range(par_ref_levels):
    pmesh.UniformRefinement()
pmesh.ReorientTetMesh()

# 6. Define the trial, interfacial (trace) and test DPG spaces:
#    - The trial space, x0_space, contains the non-interfacial unknowns and
#      has the essential BC.
#    - The interfacial space, xhat_space, contains the interfacial unknowns
#      and does not have essential BC.
#    - The test space, test_space, is an enriched space where the enrichment
#      degree may depend on the spatial dimension of the domain, the type of
#      the mesh and the trial space order.
trial_order = order
trace_order = order - 1
test_order = order  # reduced order, full order is (order + dim - 1)

if (dim == 2 and (order % 2 == 0 or (pmesh.MeshGenerator() & 2 and order > 1))):
    test_order = test_order + 1
if (test_order < trial_order):
    if myid == 0:
        print("Warning, test space not enriched enough to handle primal trial space")

x0_fec = mfem.H1_FECollection(trial_order, dim)
xhat_fec = mfem.RT_Trace_FECollection(trace_order, dim)
test_fec = mfem.L2_FECollection(test_order, dim)

x0_space = mfem.ParFiniteElementSpace(pmesh, x0_fec)
xhat_space = mfem.ParFiniteElementSpace(pmesh, xhat_fec)
test_space = mfem.ParFiniteElementSpace(pmesh, test_fec)

glob_true_s0 = x0_space.GlobalTrueVSize()
glob_true_s1 = xhat_space.GlobalTrueVSize()
glob_true_s_test = test_space.GlobalTrueVSize()

if myid == 0:
    print('\n'.join(["nNumber of Unknowns",
                     " Trial space,     X0   : " + str(glob_true_s0) +
                     " (order " + str(trial_order) + ")",
                     " Interface space, Xhat : " + str(glob_true_s1) +
                     " (order " + str(trace_order) + ")",
                     " Test space,      Y    : " + str(glob_true_s_test) +
                     " (order " + str(test_order) + ")"]))

# 7. Set up the linear form F(.) which corresponds to the right-hand side of
#    the FEM linear system, which in this case is (f,phi_i) where f=1.0 and
#    phi_i are the basis functions in the test finite element fespace.
one = mfem.ConstantCoefficient(1.0)
F = mfem.ParLinearForm(test_space)
F.AddDomainIntegrator(mfem.DomainLFIntegrator(one))
F.Assemble()

x0 = mfem.ParGridFunction(x0_space)
x0.Assign(0.0)

# 8. Set up the mixed bilinear form for the primal trial unknowns, B0,
#    the mixed bilinear form for the interfacial unknowns, Bhat,
#    the inverse stiffness matrix on the discontinuous test space, Sinv,
#    and the stiffness matrix on the continuous trial space, S0.
ess_bdr = mfem.intArray(pmesh.bdr_attributes.Max())
ess_bdr.Assign(1)
ess_dof = mfem.intArray()
x0_space.GetEssentialVDofs(ess_bdr, ess_dof)

B0 = mfem.ParMixedBilinearForm(x0_space, test_space)
B0.AddDomainIntegrator(mfem.DiffusionIntegrator(one))
B0.Assemble()
B0.EliminateEssentialBCFromTrialDofs(ess_dof, x0, F)
B0.Finalize()

Bhat = mfem.ParMixedBilinearForm(xhat_space, test_space)
Bhat.AddTraceFaceIntegrator(mfem.TraceJumpIntegrator())
Bhat.Assemble()
Bhat.Finalize()

Sinv = mfem.ParBilinearForm(test_space)
Sum = mfem.SumIntegrator()
Sum.AddIntegrator(mfem.DiffusionIntegrator(one))
Sum.AddIntegrator(mfem.MassIntegrator(one))
Sinv.AddDomainIntegrator(mfem.InverseIntegrator(Sum))
Sinv.Assemble()
Sinv.Finalize()

S0 = mfem.ParBilinearForm(x0_space)
S0.AddDomainIntegrator(mfem.DiffusionIntegrator(one))
S0.Assemble()
S0.EliminateEssentialBC(ess_bdr)
S0.Finalize()

matB0 = B0.ParallelAssemble()
del B0
matBhat = Bhat.ParallelAssemble()
del Bhat
matSinv = Sinv.ParallelAssemble()
del Sinv
matS0 = S0.ParallelAssemble()
del S0

# 9. Define the block structure of the problem, by creating the offset
#    variables. Also allocate two BlockVector objects to store the solution
#    and rhs.
x0_var = 0
xhat_var = 1
NVAR = 2  # enum in C

true_s0 = x0_space.TrueVSize()
true_s1 = xhat_space.TrueVSize()
true_s_test = test_space.TrueVSize()


true_offsets = mfem.intArray([0, true_s0, true_s0+true_s1])
true_offsets_test = mfem.intArray([0, true_s_test])


x = mfem.BlockVector(true_offsets)
b = mfem.BlockVector(true_offsets)
x.Assign(0.0)
b.Assign(0.0)

# 10. Set up the 1x2 block Least Squares DPG operator, B = [B0 Bhat],
#     the normal equation operator, A = B^t Sinv B, and
#     the normal equation right-hand-size, b = B^t Sinv F.

B = mfem.BlockOperator(true_offsets_test, true_offsets)
B.SetBlock(0, 0, matB0)
B.SetBlock(0, 1, matBhat)

A = mfem.RAPOperator(B, matSinv, B)

trueF = F.ParallelAssemble()

SinvF = mfem.HypreParVector(test_space)
matSinv.Mult(trueF, SinvF)
B.MultTranspose(SinvF, b)

# 11. Set up a block-diagonal preconditioner for the 2x2 normal equation
#
#        [ S0^{-1}     0     ]
#        [   0     Shat^{-1} ]      Shat = (Bhat^T Sinv Bhat)
#
#     corresponding to the primal (x0) and interfacial (xhat) unknowns.
#     Since the Shat operator is equivalent to an H(div) matrix reduced to
#     the interfacial skeleton, we approximate its inverse with one V-cycle
#     of the ADS preconditioner from the hypre library (in 2D we use AMS for
#     the rotated H(curl) problem).
S0inv = mfem.HypreBoomerAMG(matS0)
S0inv.SetPrintLevel(0)

Shat = mfem.RAP(matSinv, matBhat)
if (dim == 2):
    Shatinv = mfem.HypreAMS(Shat, xhat_space)
else:
    Shatinv = mfem.HypreADS(Shat, xhat_space)

P = mfem.BlockDiagonalPreconditioner(true_offsets)
P.SetDiagonalBlock(0, S0inv)
P.SetDiagonalBlock(1, Shatinv)

# 12. Solve the normal equation system using the PCG iterative solver.
#     Check the weighted norm of residual for the DPG least square problem.
#     Wrap the primal variable in a GridFunction for visualization purposes.
pcg = mfem.CGSolver(MPI.COMM_WORLD)
pcg.SetOperator(A)
pcg.SetPreconditioner(P)
pcg.SetRelTol(1e-6)
pcg.SetMaxIter(200)
pcg.SetPrintLevel(1)
pcg.Mult(b, x)

LSres = mfem.HypreParVector(test_space)
tmp = mfem.HypreParVector(test_space)

B.Mult(x, LSres)
LSres -= trueF
matSinv.Mult(LSres, tmp)

res = sqrt(mfem.InnerProduct(LSres, tmp))

if (myid == 0):
    print("\n|| B0*x0 + Bhat*xhat - F ||_{S^-1} = " + str(res))


x0.Distribute(x.GetBlock(x0_var))

# 13. Save the refined mesh and the solution in parallel. This output can
#     be viewed later using GLVis: "glvis -np <np> -m mesh -g sol".
smyid = '{:0>6d}'.format(myid)
mesh_name = "mesh."+smyid
sol_name = "sol."+smyid
pmesh.Print(mesh_name, 8)
x0.Save(sol_name, 8)

# 14. Send the solution by socket to a GLVis server.
if visualization:
    sol_sock = mfem.socketstream("localhost", 19916)
    sol_sock.send_text("parallel " + str(num_procs) + " " + str(myid))
    sol_sock.precision(8)
    sol_sock.send_solution(pmesh, x0)



# ----------------------------------------------------------------------------------------
# Routines for PyMFEM Wrapper Generation/Compile
# ----------------------------------------------------------------------------------------
import sys
import os
import re
import subprocess

__all__ = ["write_setup_local", "generate_wrapper",
           "clean_wrapper", "make_mfem_wrapper"]

from build_utils import *
from build_consts import *
import build_globals as bglb


def write_setup_local():
    '''
    create setup_local.py. parameters written here will be read
    by setup.py in mfem._ser and mfem._par
    '''
    mfemser = bglb.mfems_prefix
    mfempar = bglb.mfemp_prefix

    hyprelibpath = os.path.dirname(
        find_libpath_from_prefix('HYPRE', bglb.hypre_prefix))
    metislibpath = os.path.dirname(
        find_libpath_from_prefix('metis', bglb.metis_prefix))

    mfems_tpl = read_mfem_tplflags(bglb.mfems_prefix)
    mfemp_tpl = read_mfem_tplflags(
        bglb.mfemp_prefix) if bglb.build_parallel else ''

    print(mfems_tpl, mfemp_tpl)

    params = {'cxx_ser': bglb.cxx_command,
              'cc_ser': bglb.cc_command,
              'cxx_par': bglb.mpicxx_command,
              'cc_par': bglb.mpicc_command,
              'whole_archive': '--whole-archive',
              'no_whole_archive': '--no-whole-archive',
              'nocompactunwind': '',
              'swigflag': '-Wall -c++ -python -fastproxy -olddefs -keyword',
              'hypreinc': os.path.join(bglb.hypre_prefix, 'include'),
              'hyprelib': hyprelibpath,
              'metisinc': os.path.join(bglb.metis_prefix, 'include'),
              'metis5lib': metislibpath,
              'numpyinc': get_numpy_inc(),
              'mpi4pyinc': '',
              'mpiinc':bglb.mpiinc,
              'mfem_outside': '1' if  bglb.mfem_outside else '0',
              'mfembuilddir': os.path.join(mfempar, 'include'),
              'mfemincdir': os.path.join(mfempar, 'include', 'mfem'),
              'mfemlnkdir': os.path.join(mfempar, 'lib'),
              'mfemserbuilddir': os.path.join(mfemser, 'include'),
              'mfemserincdir': os.path.join(mfemser, 'include', 'mfem'),
              'mfemserlnkdir': os.path.join(mfemser, 'lib'),
              'mfemsrcdir': os.path.join(bglb.mfem_source),
              'mfemstpl': mfems_tpl,
              'mfemptpl': mfemp_tpl,
              'add_pumi': '',
              'add_strumpack': '',
              'add_cuda': '',
              'add_libceed': '',
              'add_suitesparse': '',
              'add_gslib': '',
              'add_gslibp': '',
              'add_gslibs': '',
              'libceedinc': os.path.join(bglb.libceed_prefix, 'include'),
              'gslibsinc': os.path.join(bglb.gslibs_prefix, 'include'),
              'gslibpinc': os.path.join(bglb.gslibp_prefix, 'include'),
              'cxxstdflag': bglb.cxxstd_flag,
              'build_mfem': '1' if bglb.build_mfem else '0',
              'bdist_wheel_dir': bglb.bdist_wheel_dir,
              }

    if bglb.build_parallel:
        params['mpi4pyinc'] = get_mpi4py_inc()

    def add_extra(xxx, inc_sub=None):
        params['add_' + xxx] = '1'
        ex_prefix = getattr(bglb, xxx + '_prefix')
        if inc_sub is None:
            params[xxx +
                   'inc'] = os.path.join(ex_prefix, 'include')
        else:
            params[xxx +
                   'inc'] = os.path.join(ex_prefix, 'include', inc_sub)

        params[xxx + 'lib'] = os.path.join(ex_prefix, 'lib')

    if bglb.enable_pumi:
        add_extra('pumi')
    if bglb.enable_strumpack:
        add_extra('strumpack')
    if bglb.enable_cuda:
        add_extra('cuda')
    if bglb.enable_libceed:
        add_extra('libceed')
    if bglb.enable_suitesparse:
        add_extra('suitesparse', inc_sub='suitesparse')
    if bglb.enable_gslib:
        add_extra('gslibs')
    if bglb.enable_gslib:
        add_extra('gslibp')

    pwd = chdir(rootdir)

    fid = open('setup_local.py', 'w')
    fid.write("#  setup_local.py \n")
    fid.write("#  generated from setup.py\n")
    fid.write("#  do not edit this directly\n")

    for key, value in params.items():
        text = key.lower() + ' = "' + value + '"'
        fid.write(text + "\n")
    fid.close()

    os.chdir(pwd)


def generate_wrapper(do_parallel):
    '''
    run swig.
    '''
    # this should work as far as we are in the same directory ?
    from multiprocessing import Pool, cpu_count
    import build_globals as bglb

    if bglb.dry_run or bglb.verbose:
        print("generating SWIG wrapper")
        print("using MFEM source", os.path.abspath(bglb.mfem_source))
    if not os.path.exists(os.path.abspath(bglb.mfem_source)):
        assert False, "MFEM source directory. Use --mfem-source=<path>"

    def ifiles():
        ifiles = os.listdir()
        ifiles = [x for x in ifiles if x.endswith('.i')]
        ifiles = [x for x in ifiles if not x.startswith('#')]
        ifiles = [x for x in ifiles if not x.startswith('.')]
        return ifiles

    def check_new(ifile):
        wfile = ifile[:-2] + '_wrap.cxx'
        if not os.path.exists(wfile):
            return True
        return os.path.getmtime(ifile) > os.path.getmtime(wfile)

    def update_integrator_exts():
        pwd = chdir(os.path.join(rootdir, 'mfem', 'common'))
        command1 = [sys.executable, "generate_lininteg_ext.py"]
        command2 = [sys.executable, "generate_bilininteg_ext.py"]
        make_call(command1)
        make_call(command2)
        os.chdir(pwd)

    def update_header_exists(mfem_source):
        print("updating the list of existing headers")
        list_of_headers = []
        L = len(mfem_source.split(os.sep))
        for (dirpath, dirnames, filenames) in os.walk(mfem_source):
            for filename in filenames:
                if filename.endswith('.hpp'):
                    dirs = dirpath.split(os.sep)[L:]
                    dirs.append(filename[:-4])
                    tmp = '_'.join(dirs)
                    xx = re.split('_|-', tmp)
                    new_name = 'FILE_EXISTS_'+'_'.join([x.upper() for x in xx])
                    if new_name not in list_of_headers:
                        list_of_headers.append(new_name)

        pwd = chdir(os.path.join(rootdir, 'mfem', 'common'))
        fid = open('existing_mfem_headers.i', 'w')
        for x in list_of_headers:
            fid.write("#define " + x + "\n")
        fid.close()
        os.chdir(pwd)

    mfemser = bglb.mfems_prefix
    mfempar = bglb.mfemp_prefix

    update_header_exists(bglb.mfem_source)

    swigflag = '-Wall -c++ -python -fastproxy -olddefs -keyword'.split(' ')

    pwd = chdir(os.path.join(rootdir, 'mfem', '_ser'))

    serflag = ['-I' + os.path.join(mfemser, 'include'),
               '-I' + os.path.join(mfemser, 'include', 'mfem'),
               '-I' + os.path.abspath(bglb.mfem_source)]
    if bglb.enable_suitesparse:
        serflag.append('-I' + os.path.join(bglb.suitesparse_prefix,
                                           'include', 'suitesparse'))

    for filename in ['lininteg.i', 'bilininteg.i']:
        command = [swig_command] + swigflag + serflag + [filename]
        make_call(command)
    update_integrator_exts()

    commands = []
    for filename in ifiles():
        if not check_new(filename):
            continue
        command = [swig_command] + swigflag + serflag + [filename]
        commands.append(command)

    mp_pool = Pool(max((cpu_count() - 1, 1)))
    with mp_pool:
        mp_pool.map(subprocess.run, commands)

    if not do_parallel:
        os.chdir(pwd)
        return

    chdir(os.path.join(rootdir, 'mfem', '_par'))

    parflag = ['-I' + os.path.join(mfempar, 'include'),
               '-I' + os.path.join(mfempar, 'include', 'mfem'),
               '-I' + os.path.abspath(bglb.mfem_source),
               '-I' + os.path.join(bglb.hypre_prefix, 'include'),
               '-I' + os.path.join(bglb.metis_prefix, 'include'),
               '-I' + get_mpi4py_inc()]

    if bglb.enable_pumi:
        parflag.append('-I' + os.path.join(bglb.pumi_prefix, 'include'))
    if bglb.enable_strumpack:
        parflag.append('-I' + os.path.join(bglb.strumpack_prefix, 'include'))
    if bglb.enable_suitesparse:
        parflag.append('-I' + os.path.join(bglb.suitesparse_prefix,
                                           'include', 'suitesparse'))

    commands = []
    for filename in ifiles():
        if filename == 'strumpack.i' and not bglb.enable_strumpack:
            continue
        if not check_new(filename):
            continue
        command = [swig_command] + swigflag + parflag + [filename]
        commands.append(command)

    mp_pool = Pool(max((cpu_count() - 1, 1)))
    with mp_pool:
        mp_pool.map(subprocess.run, commands)

    os.chdir(pwd)


def clean_wrapper():
    from pathlib import Path

    # serial
    pwd = chdir(os.path.join(rootdir, 'mfem', '_ser'))
    wfiles = [x for x in os.listdir() if x.endswith('_wrap.cxx')]

    print(os.getcwd(), wfiles)
    remove_files(wfiles)

    wfiles = [x for x in os.listdir() if x.endswith('_wrap.h')]
    remove_files(wfiles)

    wfiles = [x for x in os.listdir() if x.endswith('.py')]
    wfiles.remove("__init__.py")
    wfiles.remove("setup.py")
    wfiles.remove("tmop_modules.py")
    remove_files(wfiles)

    ifiles = [x for x in os.listdir() if x.endswith('.i')]
    for x in ifiles:
        Path(x).touch()

    # parallel
    chdir(os.path.join(rootdir, 'mfem', '_par'))
    wfiles = [x for x in os.listdir() if x.endswith('_wrap.cxx')]

    remove_files(wfiles)
    wfiles = [x for x in os.listdir() if x.endswith('_wrap.h')]
    remove_files(wfiles)

    wfiles = [x for x in os.listdir() if x.endswith('.py')]
    wfiles.remove("__init__.py")
    wfiles.remove("setup.py")
    wfiles.remove("tmop_modules.py")
    remove_files(wfiles)

    ifiles = [x for x in os.listdir() if x.endswith('.i')]
    for x in ifiles:
        Path(x).touch()

    chdir(pwd)


def make_mfem_wrapper(serial=True):
    '''
    compile PyMFEM wrapper code
    '''
    from multiprocessing import cpu_count
    import build_globals as bglb

    if bglb.dry_run or bglb.verbose:
        print("compiling wrapper code, serial=" + str(serial))
    if not os.path.exists(os.path.abspath(bglb.mfem_source)):
        assert False, "MFEM source directory. Use --mfem-source=<path>"

    record_mfem_sha(bglb.mfem_source)

    write_setup_local()

    if serial:
        pwd = chdir(os.path.join(rootdir, 'mfem', '_ser'))
    else:
        pwd = chdir(os.path.join(rootdir, 'mfem', '_par'))

    python = sys.executable
    command = [python, 'setup.py', 'build_ext', '--inplace', '--parallel',
               str(max((cpu_count() - 1, 1)))]
    make_call(command, force_verbose=True)

    os.chdir(pwd)




'''
   MFEM example 25
      See c++ version in the MFEM library for more detail

      This example also demonstrates how to use JITed function in
      more genral form.

      In particular it shows.
         * mfem.jit.vector decorator as a function call
         * defining constant parameters in functions using a params keyword
         * calling JITed function from JITed coefficient

'''
from numba import jit, types, carray
import os
import mfem.ser as mfem
from mfem.ser import intArray
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sin, cos, exp, sqrt, pi

try:
    import numba
except ImportError:
    assert False, "Numba is required"

from mfem.common.bessel import jv as jn
from mfem.common.bessel import yv as yn

prob = ''


def run(meshfile="",
        order=1,
        ref_levels=0,
        visualization=1,
        herm_conv=True,
        device_config='cpu',
        pa=False):

    # 2. Enable hardware devices such as GPUs, and programming models such as
    #    CUDA, OCCA, RAJA and OpenMP based on command line options.
    device = mfem.Device(device_config)
    device.Print()

    # 3. Setup the mesh
    if meshfile == '':
        exact_known = True
        if prob == "beam":
            meshfile = "beam-hex.mesh"
        elif prob == "disc":
            meshfile = "square-disc.mesh"
        elif prob == "lshape":
            meshfile = "l-shape.mesh"
        elif prob == "fichera":
            meshfile = "fichera.mesh"
        else:
            meshfile = "inline-quad.mesh"
            exact_known = False
    else:
        exact_known = True

    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', meshfile))

    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()

    # Setup PML length
    length = np.zeros((dim, 2))

    # 4. Setup the Cartesian PML region.
    if prob == "disc":
        length[:] = 0.2
    elif prob == "lshape":
        length[0, 0] = 0.1
        length[1, 0] = 0.1
    elif prob == "fichera":
        length[0, 1] = 0.5
        length[1, 1] = 0.5
        length[2, 1] = 0.5
    elif prob == "beam":
        length[0, 1] = 2.0
    else:
        length[:] = 0.25

    pml = CartesianPML(mesh, length)
    comp_domain_bdr = pml.comp_dom_bdr
    domain_bdr = pml.dom_bdr

    # 5. Refine the mesh to increase the resolution.
    for l in range(ref_levels):
        mesh.UniformRefinement()

    # 6. Set element attributes in order to distinguish elements in the
    #    PML region
    pml.SetAttributes(mesh)

    # 7. Define a finite element space on the mesh. Here we use the Nedelec
    #    finite elements of the specified order.
    fec = mfem.ND_FECollection(order, dim)
    fespace = mfem.FiniteElementSpace(mesh, fec)

    size = fespace.GetTrueVSize()

    print("Number of finite element unknowns: " + str(size))

    # 8. Determine the list of true essential boundary dofs. In this example,
    #    the boundary conditions are defined based on the specific mesh and the
    #    problem type.

    battrs = mesh.GetBdrAttributeArray()

    if len(battrs) > 0:
        if prob == "lshape" or prob == "fichera":
            ess_bdr0 = [0]*np.max(battrs)

            for j in range(mesh.GetNBE()):
                bdrgeom = mesh.GetBdrElementBaseGeometry(j)
                tr = mesh.GetBdrElementTransformation(j)
                center = tr.Transform(mfem.Geometries.GetCenter(bdrgeom))

                k = mesh.GetBdrAttribute(j)
                if prob == "lshape":
                    if (center[0] == 1.0 or center[0] == 0.5 or
                            center[1] == 0.5):
                        ess_bdr0[k - 1] = 1
                else:  # prob == "fichera"
                    if (center[0] == -1.0 or center[0] == 0.0 or
                            center[1] == 0.0 or center[2] == 0.0):
                        ess_bdr0[k - 1] = 1
        else:
            ess_bdr0 = [1]*np.max(battrs)

        ess_bdr = mfem.intArray(ess_bdr0)
    else:
        ess_bdr = mfem.intArray()

    ess_tdof_list = mfem.intArray()
    fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

    # 9. Setup Complex Operator convention
    conv = herm_conv if mfem.ComplexOperator.HERMITIAN else mfem.ComplexOperator.BLOCK_SYMMETRIC

    # 10. Set up the linear form b(.) which corresponds to the right-hand side of
    #     the FEM linear system.
    params = {"comp_domain_bdr": comp_domain_bdr,
              "dim": dim,
              "omega": omega,
              "epsilon": epsilon,
              "prob": prob,
              "mu": mu}
    f = mfem.jit.vector(vdim=dim,
                        interface="c++",
                        params=params)(source)
    b = mfem.ComplexLinearForm(fespace, conv)
    if prob == "general":
        b.AddDomainIntegrator(None, mfem.VectorFEDomainLFIntegrator(f))

    b.Assign(0.0)
    b.Assemble()

    # 11. Define the solution vector x as a complex finite element grid function
    #     corresponding to fespace.
    x = mfem.ComplexGridFunction(fespace)
    x.Assign(0.0)

    sig = types.void(types.double[:],
                     types.complex128[:], types.intc)
    exact_solution = mfem.jit.func(sig, params=params)(maxwell_solution)

    params = {'comp_domain_bdr': comp_domain_bdr,
              'exact_solution': exact_solution}
    E_Re = mfem.jit.vector(vdim=dim, interface="c++",
                           params=params)(E_bdr_data_Re)
    E_Im = mfem.jit.vector(vdim=dim, interface="c++",
                           params=params)(E_bdr_data_Im)
    x.ProjectBdrCoefficientTangent(E_Re, E_Im, ess_bdr)

    # 12. Set up the sesquilinear form a(.,.)
    #
    #     In Comp
    #     Domain:   1/mu (Curl E, Curl F) - omega^2 * epsilon (E,F)
    #
    #     In PML:   1/mu (1/det(J) J^T J Curl E, Curl F)
    #               - omega^2 * epsilon (det(J) * (J^T J)^-1 * E, F)
    #
    #     where J denotes the Jacobian Matrix of the PML Stretching function

    attrs = mesh.GetAttributeArray()
    if len(attrs) > 0:
        attr = [0]*np.max(attrs)
        attrPML = [0]*np.max(attrs)

        attr[0] = 1
        if max(attrs) > 1:
            attrPML[1] = 1

    muinv = mfem.ConstantCoefficient(1/mu)
    omeg = mfem.ConstantCoefficient(-omega**2 * epsilon)
    attr = mfem.intArray(attr)
    attrPML = mfem.intArray(attrPML)
    restr_muinv = mfem.RestrictedCoefficient(muinv, attr)
    restr_omeg = mfem.RestrictedCoefficient(omeg, attr)

    # Integrators inside the computational domain (excluding the PML region)
    a = mfem.SesquilinearForm(fespace, conv)
    a.AddDomainIntegrator(mfem.CurlCurlIntegrator(restr_muinv), None)
    a.AddDomainIntegrator(mfem.VectorFEMassIntegrator(restr_omeg), None)

    cdim = 1 if dim == 2 else dim

    def dm(x, m):
        sdim = len(x)
        diag = np.empty(sdim)
        diag_func(x, diag)
        for i in range(sdim):
            m[i] = diag[i]

    # JIT compiles all functions first. params defines local variables
    # inside the JITed function.
    sig = types.void(types.double[:], types.float64[:])
    params = {"StretchFunction": pml.StretchFunction, "dim": dim}
    detJ_inv_JT_J_Re = mfem.jit.func(sig, params=params)(detJ_inv_JT_J_Re_f)
    detJ_inv_JT_J_Im = mfem.jit.func(sig, params=params)(detJ_inv_JT_J_Im_f)
    detJ_inv_JT_J_abs = mfem.jit.func(sig, params=params)(detJ_inv_JT_J_abs_f)
    params = {"StretchFunction": pml.StretchFunction, "dim": dim}
    detJ_JT_J_inv_Re = mfem.jit.func(sig, params=params)(detJ_JT_J_inv_Re_f)
    detJ_JT_J_inv_Im = mfem.jit.func(sig, params=params)(detJ_JT_J_inv_Im_f)
    detJ_JT_J_inv_abs = mfem.jit.func(sig, params=params)(detJ_JT_J_inv_abs_f)

    pml_c1_Re = mfem.jit.vector(shape=(cdim, ), interface="c++",
                                params={"diag_func": detJ_inv_JT_J_Re})(dm)
    pml_c1_Im = mfem.jit.vector(shape=(cdim, ), interface="c++",
                                params={"diag_func": detJ_inv_JT_J_Im})(dm)

    c1_Re = mfem.ScalarVectorProductCoefficient(muinv, pml_c1_Re)
    c1_Im = mfem.ScalarVectorProductCoefficient(muinv, pml_c1_Im)
    restr_c1_Re = mfem.VectorRestrictedCoefficient(c1_Re, attrPML)
    restr_c1_Im = mfem.VectorRestrictedCoefficient(c1_Im, attrPML)

    pml_c2_Re = mfem.jit.vector(shape=(dim,), interface="c++",
                                params={"diag_func": detJ_JT_J_inv_Re})(dm)
    pml_c2_Im = mfem.jit.vector(shape=(dim,), interface="c++",
                                params={"diag_func": detJ_JT_J_inv_Im})(dm)
    c2_Re = mfem.ScalarVectorProductCoefficient(omeg, pml_c2_Re)
    c2_Im = mfem.ScalarVectorProductCoefficient(omeg, pml_c2_Im)
    restr_c2_Re = mfem.VectorRestrictedCoefficient(c2_Re, attrPML)
    restr_c2_Im = mfem.VectorRestrictedCoefficient(c2_Im, attrPML)

    # Integrators inside the PML region
    a.AddDomainIntegrator(mfem.CurlCurlIntegrator(restr_c1_Re),
                          mfem.CurlCurlIntegrator(restr_c1_Im))
    a.AddDomainIntegrator(mfem.VectorFEMassIntegrator(restr_c2_Re),
                          mfem.VectorFEMassIntegrator(restr_c2_Im))

    # 13. Assemble the bilinear form and the corresponding linear system,
    #     applying any necessary transformations such as: assembly, eliminating
    #     boundary conditions, applying conforming constraints for
    #     non-conforming AMR, etc.
    if pa:
        a.SetAssemblyLevel(mfem.AssemblyLevel.PARTIAL)
    a.Assemble(0)

    A = mfem.OperatorPtr()
    B = mfem.Vector()
    X = mfem.Vector()
    a.FormLinearSystem(ess_tdof_list, x, b, A, X, B)

    # 14a. Set up the Bilinear form a(.,.) for the preconditioner
    #
    #    In Comp
    #    Domain:   1/mu (Curl E, Curl F) + omega^2 * epsilon (E,F)
    #
    #    In PML:   1/mu (abs(1/det(J) J^T J) Curl E, Curl F)
    #              + omega^2 * epsilon (abs(det(J) * (J^T J)^-1) * E, F)

    umf_solver = False
    if pa or not umf_solver:
        absomeg = mfem.ConstantCoefficient(omega**2 * epsilon)
        restr_absomeg = mfem.RestrictedCoefficient(absomeg, attr)

        prec = mfem.BilinearForm(fespace)
        prec.AddDomainIntegrator(mfem.CurlCurlIntegrator(restr_muinv))
        prec.AddDomainIntegrator(mfem.VectorFEMassIntegrator(restr_absomeg))

        pml_c1_abs = mfem.jit.vector(shape=(cdim,),
                                     params={
                                         "diag_func": detJ_inv_JT_J_abs, "dim": cdim, },
                                     interface="c++")(dm)
        c1_abs = mfem.ScalarVectorProductCoefficient(muinv, pml_c1_abs)
        restr_c1_abs = mfem.VectorRestrictedCoefficient(c1_abs, attrPML)

        pml_c2_abs = mfem.jit.vector(shape=(dim,),
                                     params={
                                         "diag_func": detJ_JT_J_inv_abs, "dim": dim},
                                     interface="c++")(dm)
        c2_abs = mfem.ScalarVectorProductCoefficient(absomeg, pml_c2_abs)
        restr_c2_abs = mfem.VectorRestrictedCoefficient(c2_abs, attrPML)

        prec.AddDomainIntegrator(mfem.CurlCurlIntegrator(restr_c1_abs))
        prec.AddDomainIntegrator(mfem.VectorFEMassIntegrator(restr_c2_abs))

        if pa:
            prec.SetAssemblyLevel(mfem.AssemblyLevel.PARTIAL)
        prec.Assemble()

        # 14b. Define and apply a GMRES solver for AU=B with a block diagonal
        #      preconditioner based on the Gauss-Seidel or Jacobi sparse smoother.
        offsets = intArray([0, fespace.GetTrueVSize(), fespace.GetTrueVSize()])
        offsets.PartialSum()

        s = -1.0 if conv == mfem.ComplexOperator.HERMITIAN else 1.0
        if pa:
            # Jacobi Smoother
            d00 = mfem.OperatorJacobiSmoother(prec, ess_tdof_list)
            d11 = mfem.ScaledOperator(d00, s)
            pc_r = d00
            pc_i = d11
        else:
            PCOpAh = mfem.OperatorPtr()
            prec.SetDiagonalPolicy(mfem.Operator.DIAG_ONE)
            prec.FormSystemMatrix(ess_tdof_list, PCOpAh)

            # Gauss-Seidel Smoother
            gs00 = mfem.GSSmoother(PCOpAh.AsSparseMatrix())
            gs11 = mfem.ScaledOperator(gs00, s)
            pc_r = gs00
            pc_i = gs11

        BlockDP = mfem.BlockDiagonalPreconditioner(offsets)
        BlockDP.SetDiagonalBlock(0, pc_r)
        BlockDP.SetDiagonalBlock(1, pc_i)

        gmres = mfem.GMRESSolver()
        gmres.SetPrintLevel(1)
        gmres.SetKDim(200)
        gmres.SetMaxIter(5000 if pa else 2000)
        # gmres.SetMaxIter(1)
        gmres.SetRelTol(1e-5)
        gmres.SetAbsTol(0.0)
        gmres.SetOperator(A.Ptr())
        gmres.SetPreconditioner(BlockDP)
        gmres.Mult(B, X)

    # 15. Recover the solution as a finite element grid function and compute the
    #     errors if the exact solution is known.
    a.RecoverFEMSolution(X, b, x)

    # If exact is known compute the error
    if exact_known:
        E_ex_Re = mfem.jit.vector(
            params={"exact_solution": exact_solution, "sdim": dim})(E_exact_Re)
        E_ex_Im = mfem.jit.vector(
            params={"exact_solution": exact_solution, "sdim": dim})(E_exact_Im)

        order_quad = max([2, 2 * order + 1])

        birs = [mfem.IntRules.Get(i, order_quad)
                for i in range(mfem.Geometry.NumGeom)]

        L2Error_Re = x.real().ComputeL2Error(E_ex_Re, birs, pml.elems)
        L2Error_Im = x.imag().ComputeL2Error(E_ex_Im, birs, pml.elems)

        x_gf0 = mfem.ComplexGridFunction(fespace)
        x_gf0.Assign(0.0)
        norm_E_Re = x_gf0.real().ComputeL2Error(E_ex_Re, birs, pml.elems)
        norm_E_Im = x_gf0.imag().ComputeL2Error(E_ex_Im, birs, pml.elems)

        print("")
        print(" Relative Error (Re part): || E_h - E || / ||E|| = " +
              "{:g}".format(L2Error_Re / norm_E_Re))
        print(" Relative Error (Im part): || E_h - E || / ||E|| = " +
              "{:g}".format(L2Error_Im / norm_E_Im))
        print(" Total Error : " +
              "{:g}".format(sqrt(L2Error_Re*L2Error_Re + L2Error_Im*L2Error_Im)))
        print("")

    mesh.Print("ex25.mesh", 8)
    x.real().Save("ex25-sol_r.gf", 8)
    x.imag().Save("ex25-sol_i.gf", 8)

    if visualization:
        keys = "keys macF\n" if dim == 3 else "keys amrRljcUUuu\n"
        if prob == "beam" and dim == 3:
            keys = "keys macFFiYYYYYYYYYYYYYYYYYY\n"
        if prob == "beam" and dim == 2:
            keys = "keys amrRljcUUuuu\n"

        sol_sock_re = mfem.socketstream("localhost", 19916)
        sol_sock_re.precision(8)
        sol_sock_re << "solution\n" << mesh << x.real() << keys
        sol_sock_re << "window_title 'Soluiton real part'"
        sol_sock_re.flush()

        sol_sock_im = mfem.socketstream("localhost", 19916)
        sol_sock_im.precision(8)
        sol_sock_im << "solution\n" << mesh << x.imag() << keys
        sol_sock_im << "window_title 'Soluiton imag part'"
        sol_sock_im.flush()

        x_t = mfem.GridFunction(fespace)
        x_t.Assign(x.real())

        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock.precision(8)
        sol_sock << "solution\n" << mesh << x_t << keys << "autoscale off\n"
        sol_sock << "window_title 'Harmonic Solution (t = 0.0T)'"
        sol_sock << "pause\n"
        sol_sock.flush()

        print("GLVis visualization paused. Press space (in the GLVis window) to resume it.")
        num_frames = 32
        i = 0

        for i in range(num_frames):
            t = (i % num_frames) / num_frames
            oss = "Harmonic Solution (t = " + str(t) + " T)"
            dd = (cos(2.0 * pi * t)*x.real().GetDataArray() +
                  sin(2.0 * pi * t)*x.imag().GetDataArray())
            x_t.Assign(dd)
            # x_t.Assign(mfem.Vector(dd))
            sol_sock << "solution\n" << mesh << x_t
            sol_sock << "window_title '" << oss << "'"
            sol_sock.flush()


class CartesianPML:
    def __init__(self, mesh, length):
        self.length = length
        self.dim = mesh.Dimension()
        self.SetBoundaries(mesh)

    def SetBoundaries(self, mesh):
        self.comp_dom_bdr = np.zeros((self.dim, 2))
        self.dom_bdr = np.zeros((self.dim, 2))
        pmin, pmax = mesh.GetBoundingBox()
        for i in range(self.dim):
            self.dom_bdr[i, 0] = pmin[i]
            self.dom_bdr[i, 1] = pmax[i]
            self.comp_dom_bdr[i, 0] = self.dom_bdr[i, 0] + self.length[i, 0]
            self.comp_dom_bdr[i, 1] = self.dom_bdr[i, 1] - self.length[i, 1]

    def SetAttributes(self, mesh):
        # Initialize bdr attributes
        self.elems = mfem.intArray(mesh.GetNE())

        for i in range(mesh.GetNBE()):
            mesh.GetBdrElement(i).SetAttribute(i+1)

        #  Loop through the elements and identify which of them are in the PML
        for i in range(mesh.GetNE()):
            self.elems[i] = 1

            in_pml = False
            el = mesh.GetElement(i)

            #  Initialize attribute
            el.SetAttribute(1)
            vertices = el.GetVerticesArray()
            nrvert = len(vertices)

            # Check if any vertex is in the PML
            for iv in range(nrvert):
                vert_idx = vertices[iv]
                coords = mesh.GetVertexArray(vert_idx)

                for comp in range(self.dim):
                    if (coords[comp] > self.comp_dom_bdr[comp, 1] or
                            coords[comp] < self.comp_dom_bdr[comp, 0]):
                        in_pml = True
                        break

            if in_pml:
                self.elems[i] = 0
                el.SetAttribute(2)

        # construct attribute array in Mesh object
        mesh.SetAttributes()
        self.StretchFunction = self._GenerateStretchFunction()

    def _GenerateStretchFunction(self):
        sig = types.void(types.double[:], types.complex128[:])
        params = {"comp_domain_bdr": self.comp_dom_bdr,
                  "dim": self.dim,
                  "length": self.length,
                  "omega": omega,
                  "epsilon": epsilon,
                  "mu": mu}

        def _StretchFunction(x, dxs):
            zi = 1j

            n = 2.0
            c = 5.0
            k = omega * sqrt(epsilon * mu)

            # Stretch in each direction independently
            for i in range(dim):
                dxs[i] = 1.0
                if x[i] >= comp_domain_bdr[i, 1]:
                    coeff = n * c / k / length[i, 1]**n
                    dxs[i] = (1.0 + zi * coeff *
                              abs((x[i] - comp_domain_bdr[i, 1])**(n-1.0)))
                if x[i] <= comp_domain_bdr[i, 0]:
                    coeff = n * c / k / length[i, 0]**n
                    dxs[i] = (1.0 + zi * coeff *
                              abs((x[i] - comp_domain_bdr[i, 0])**(n-1.0)))

        func = mfem.jit.func(sig, params=params)(_StretchFunction)
        return func

#
#  functions (these are JITed using Numba insdie run())
#


def source(x, out):
    center = np.zeros(dim)
    r = 0
    for i in range(dim):
        center[i] = 0.5 * (comp_domain_bdr[i, 0] + comp_domain_bdr[i, 1])
        r += (x[i] - center[i])**2
        out[i] = 0

    n = 5.0 * omega * sqrt(epsilon * mu) / pi
    coeff = n**2 / pi
    alpha = -n**2 * r

    out[0] = coeff * exp(alpha)


def maxwell_solution(x, E, sdim):
    # Initialize
    for i in range(sdim):
        E[i] = 0.0
    zi = 1j
    k = omega * sqrt(epsilon * mu)

    if prob == "disc" or prob == "lshape" or prob == "fichera":
        shift = np.zeros(sdim)
        if prob == "fichera":
            shift += 1.0
        elif prob == "disc":
            shift -= 0.5
        else:
            shift -= 1.0

        if sdim == 2:
            x0 = x[0] + shift[0]
            x1 = x[1] + shift[1]
            r = sqrt(x0 * x0 + x1 * x1)
            beta = k * r

            # Bessel functions
            Ho = jn(0.0, beta) + zi * yn(0, beta)
            Ho_r = -k * (jn(1.0, beta) + zi * yn(1, beta))
            Ho_rr = -k * k * (1.0 / beta *
                              (jn(1., beta) + zi * yn(1, beta)) -
                              (jn(2., beta) + zi * yn(2, beta)))

            # First derivatives
            r_x = x0 / r
            r_y = x1 / r
            r_xy = -(r_x / r) * r_y
            r_xx = (1.0 / r) * (1.0 - r_x * r_x)

            val = 0.25 * zi * Ho
            val_xx = 0.25 * zi * (r_xx * Ho_r + r_x * r_x * Ho_rr)
            val_xy = 0.25 * zi * (r_xy * Ho_r + r_x * r_y * Ho_rr)
            E[0] = zi / k * (k * k * val + val_xx)
            E[1] = zi / k * val_xy
        elif sdim == 3:
            x0 = x[0] + shift[0]
            x1 = x[1] + shift[1]
            x2 = x[2] + shift[2]
            r = sqrt(x0 * x0 + x1 * x1 + x2 * x2)

            r_x = x0 / r
            r_y = x1 / r
            r_z = x2 / r

            r_xx = (1.0 / r) * (1.0 - r_x * r_x)
            r_yx = -(r_y / r) * r_x
            r_zx = -(r_z / r) * r_x

            val = exp(zi * k * r) / r
            val_r = val / r * (zi * k * r - 1.0)
            val_rr = val / (r * r) * (-k * k * r * r
                                      - 2.0 * zi * k * r + 2.0)

            val_xx = val_rr * r_x * r_x + val_r * r_xx
            val_yx = val_rr * r_x * r_y + val_r * r_yx
            val_zx = val_rr * r_x * r_z + val_r * r_zx

            alpha = zi * k / 4.0 / pi / k / k

            E[0] = alpha * (k * k * val + val_xx)
            E[1] = alpha * val_yx
            E[2] = alpha * val_zx
        else:
            pass
    elif prob == 'beam':
        # T_10 mode
        if sdim == 3:
            k10 = sqrt(k * k - pi*pi)
            E[1] = -zi * k / pi * sin(pi*x[2])*exp(zi * k10 * x[0])
        elif sdim == 2:
            E[1] = -zi * k / pi * exp(zi * k * x[0])
        else:
            pass
    else:
        pass


def E_exact_Re(x, E):
    sdim = len(x)
    E_ = np.empty(sdim, dtype=np.complex128)
    exact_solution(x, E_, sdim)
    for i in range(sdim):
        E[i] = E_[i].real


def E_exact_Im(x, E):
    sdim = len(x)
    E_ = np.empty(sdim, dtype=np.complex128)
    exact_solution(x, E_, sdim)
    for i in range(sdim):
        E[i] = E_[i].imag


def E_bdr_data_Re(x, E):
    sdim = len(x)
    for i in range(sdim):
        E[i] = 0.0
    in_pml = False

    for i in range(sdim):
        # check if in PML
        if ((x[i] - comp_domain_bdr[i, 0]) < 0.0 or
                (x[i] - comp_domain_bdr[i, 1]) > 0.0):
            in_pml = True
            break
    if not in_pml:
        E_ = np.empty(sdim, dtype=np.complex128)
        exact_solution(x, E_, sdim)
        for i in range(sdim):
            E[i] = E_[i].real


def E_bdr_data_Im(x, E):
    sdim = len(x)
    for i in range(sdim):
        E[i] = 0.0
    in_pml = False

    for i in range(sdim):
        # check if in PML
        if ((x[i] - comp_domain_bdr[i, 0]) < 0.0 or
                (x[i] - comp_domain_bdr[i, 1]) > 0.0):
            in_pml = True
            break
    if not in_pml:
        E_ = np.empty(sdim, dtype=np.complex128)
        exact_solution(x, E_, sdim)
        for i in range(sdim):
            E[i] = E_[i].imag


def detJ_JT_J_inv_Re_f(x, D):
    dxs = np.empty(dim, dtype=np.complex128)
    det = complex(1.0)
    StretchFunction(x, dxs)
    for i in range(dim):
        det *= dxs[i]
    for i in range(dim):
        D[i] = (det / (dxs[i]**2)).real
    # print(D)


def detJ_JT_J_inv_Im_f(x, D):
    dxs = np.empty(dim, dtype=np.complex128)
    det = complex(1.0)
    StretchFunction(x, dxs)
    for i in range(dim):
        det *= dxs[i]
    for i in range(dim):
        D[i] = (det / (dxs[i]**2)).imag
    # print("imag", D)


def detJ_JT_J_inv_abs_f(x, D):
    dxs = np.empty(dim, dtype=np.complex128)
    det = complex(1.0)
    StretchFunction(x, dxs)
    for i in range(dim):
        det *= dxs[i]
    for i in range(dim):
        D[i] = abs(det / (dxs[i]**2))


def detJ_inv_JT_J_Re_f(x, D):
    dxs = np.empty(dim, dtype=np.complex128)
    det = 1.0
    StretchFunction(x, dxs)
    for i in range(dim):
        det *= dxs[i]
    # in the 2D case the coefficient is scalar 1/det(J)
    if dim == 2:
        D[0] = (1.0 / det).real
    else:
        for i in range(dim):
            D[i] = (dxs[i]**2 / det).real


def detJ_inv_JT_J_Im_f(x, D):
    dxs = np.empty(dim, dtype=np.complex128)
    det = 1.0
    StretchFunction(x, dxs)
    for i in range(dim):
        det *= dxs[i]
    # in the 2D case the coefficient is scalar 1/det(J)
    if dim == 2:
        D[0] = (1.0 / det).imag
    else:
        for i in range(dim):
            D[i] = (dxs[i]**2 / det).imag


def detJ_inv_JT_J_abs_f(x, D):
    dxs = np.empty(dim, dtype=np.complex128)
    det = 1.0
    StretchFunction(x, dxs)
    for i in range(dim):
        det *= dxs[i]
    # in the 2D case the coefficient is scalar 1/det(J)
    if dim == 2:
        D[0] = abs(1.0 / det)
    else:
        for i in range(dim):
            D[i] = abs(dxs[i]**2 / det)


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(
        description='Ex25 (PML)')
    parser.add_argument('-m', '--mesh',
                        default="",
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree)")
    parser.add_argument("-prob",
                        "--problem-type",
                        action='store', type=int, default=4,
                        help=" 0: beam, 1: disc, 2: lshape, 3: fichera, 4: General")
    parser.add_argument("-ref", "--refine",
                        action='store', type=int, default=3,
                        help="Number of times to refine the mesh uniformly.")
    parser.add_argument("-mu", "--permeability",
                        action='store', type=float, default=1.0,
                        help="Permeability of free space (or 1/(spring constant)).")
    parser.add_argument("-eps", "--permittivity",
                        action='store', type=float, default=1.0,
                        help="Permittivity of free space (or mass constant).")
    parser.add_argument("-f", "--frequency",
                        action='store',
                        type=float,
                        default=5.0,
                        help="Set the frequency for the exact")
    parser.add_argument("-no-herm", "--no-hermitian",
                        action='store_false',
                        default=True,
                        help="Do not use convention for Hermitian operators.")
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        default=True,
                        help='Enable GLVis visualization')
    parser.add_argument("-pa", "--partial-assembly",
                        action='store_true',
                        help="Enable Partial Assembly.")
    parser.add_argument("-d", "--device",
                        default="cpu", type=str,
                        help="Device configuration string, see Device::Configure().")

    args = parser.parse_args()
    parser.print_options(args)

    probs = {0: "beam", 1: "disc", 2: "lshape", 3: "fichera", 4: "general"}
    globals()["prob"] = probs[args.problem_type]
    globals()["omega"] = 2*pi*args.frequency
    globals()["epsilon"] = args.permittivity
    globals()["mu"] = args.permeability
    run(meshfile=args.mesh,
        order=args.order,
        ref_levels=args.refine,
        visualization=args.visualization,
        herm_conv=args.no_hermitian,
        device_config=args.device,
        pa=args.partial_assembly)



'''
   MFEM example 16

   How to run:
      mpirun -np 4 python <arguments>

   Example of arguments:
      ex16p.py -m inline-tri.mesh
      ex16p.py -m disc-nurbs.mesh -tf 2
      ex16p.py -s 1 -a 0.0 -k 1.0
      ex16p.py -s 2 -a 1.0 -k 0.0
      ex16p.py -s 3 -a 0.5 -k 0.5 -o 4
      ex16p.py -s 14 -dt 1.0e-4 -tf 4.0e-2 -vs 40
      ex16p.py -m fichera-q2.mesh
      ex16p.py -m escher.mesh
      ex16p.py -m beam-tet.mesh -tf 10 -dt 0.1
      ex16p.py -m amr-quad.mesh -o 4 -rs 0 -rp 0
      ex16p.py -m amr-hex.mesh -o 2 -rs 0 -rp 0
'''
import sys
from mfem.common.arg_parser import ArgParser
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sqrt, pi, cos, sin, hypot, arctan2
from scipy.special import erfc

from mfem.par import intArray, add_vector
import mfem.par as mfem
from mpi4py import MPI

num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank


class ConductionOperator(mfem.PyTimeDependentOperator):
    def __init__(self, fespace, alpha, kappa, u):
        mfem.PyTimeDependentOperator.__init__(
            self, fespace.GetTrueVSize(), 0.0)
        rel_tol = 1e-8
        self.alpha = alpha
        self.kappa = kappa
        self.T = None
        self.K = None
        self.M = None
        self.fespace = fespace

        self.ess_tdof_list = intArray()
        self.Mmat = mfem.HypreParMatrix()
        self.Kmat = mfem.HypreParMatrix()
        self.M_solver = mfem.CGSolver(fespace.GetComm())
        self.M_prec = mfem.HypreSmoother()
        self.T_solver = mfem.CGSolver(fespace.GetComm())
        self.T_prec = mfem.HypreSmoother()
        self.z = mfem.Vector(self.Height())

        self.M = mfem.ParBilinearForm(fespace)
        self.M.AddDomainIntegrator(mfem.MassIntegrator())
        self.M.Assemble()
        self.M.FormSystemMatrix(self.ess_tdof_list, self.Mmat)

        self.M_solver.iterative_mode = False
        self.M_solver.SetRelTol(rel_tol)
        self.M_solver.SetAbsTol(0.0)
        self.M_solver.SetMaxIter(100)
        self.M_solver.SetPrintLevel(0)
        self.M_prec.SetType(mfem.HypreSmoother.Jacobi)
        self.M_solver.SetPreconditioner(self.M_prec)
        self.M_solver.SetOperator(self.Mmat)

        self.T_solver.iterative_mode = False
        self.T_solver.SetRelTol(rel_tol)
        self.T_solver.SetAbsTol(0.0)
        self.T_solver.SetMaxIter(100)
        self.T_solver.SetPrintLevel(0)
        self.T_solver.SetPreconditioner(self.T_prec)

        self.SetParameters(u)

    def Mult(self, u, u_dt):
        # Compute:
        #  du_dt = M^{-1}*-K(u) for du_dt
        self.Kmat.Mult(u, self.z)
        self.z.Neg()   # z = -z
        self.M_solver.Mult(self.z, du_dt)

    def ImplicitSolve(self, dt, u, du_dt):
        # Solve the equation:
        #    du_dt = M^{-1}*[-K(u + dt*du_dt)]
        #    for du_dt
        if self.T is None:
            self.T = mfem.Add(1.0, self.Mmat, dt, self.Kmat)
            current_dt = dt
            self.T_solver.SetOperator(self.T)
        self.Kmat.Mult(u, self.z)
        self.z.Neg()
        self.T_solver.Mult(self.z, du_dt)

    def SetParameters(self, u):
        u_alpha_gf = mfem.ParGridFunction(self.fespace)
        u_alpha_gf.SetFromTrueDofs(u)
        for i in range(u_alpha_gf.Size()):
            u_alpha_gf[i] = self.kappa + self.alpha * u_alpha_gf[i]

        self.K = mfem.ParBilinearForm(self.fespace)
        u_coeff = mfem.GridFunctionCoefficient(u_alpha_gf)
        self.K.AddDomainIntegrator(mfem.DiffusionIntegrator(u_coeff))
        self.K.Assemble(0)
        self.K.FormSystemMatrix(self.ess_tdof_list, self.Kmat)
        self.T = None


class InitialTemperature(mfem.PyCoefficient):
    def EvalValue(self, x):
        xx = np.array(x)
        norm2 = np.sqrt(float(np.sum(xx**2)))
        if norm2 < 0.5:
            return 2.0
        return 1.0


parser = ArgParser(description='Ex16p')
parser.add_argument('-m', '--mesh',
                    default='star.mesh',
                    action='store', type=str,
                    help='Mesh file to use.')
parser.add_argument('-rs', '--refine-serial',
                    action='store', default=2, type=int,
                    help="Number of times to refine the mesh uniformly in serial")
parser.add_argument('-rp', '--refine-parallel',
                    action='store', default=1, type=int,
                    help="Number of times to refine the mesh uniformly in parallel")

parser.add_argument('-o', '--order',
                    action='store', default=2, type=int,
                    help="Finite element order (polynomial degree)")
parser.add_argument('-s', '--ode-solver',
                    action='store', default=3, type=int,
                    help='\n'.join(["ODE solver: 1 - Backward Euler, 2 - SDIRK2, 3 - SDIRK3",
                                    "\t\t 11 - Forward Euler, 12 - RK2, 13 - RK3 SSP, 14 - RK4."]))
parser.add_argument('-t', '--t-final',
                    action='store', default=0.5, type=float,
                    help="Final time; start time is 0.")
parser.add_argument("-dt", "--time-step",
                    action='store', default=0.01, type=float,
                    help="Time step.")
parser.add_argument('-a', '--alpha',
                    action='store', default=0.01, type=float,
                    help='Alpha coefficient')
parser.add_argument('-k', '--kappa',
                    action='store', default=0.5, type=float,
                    help='Kappa coefficient')
parser.add_argument('-vis', '--visualization',
                    action='store_true',
                    help='Enable GLVis visualization')
parser.add_argument('-vs', '--visualization-steps',
                    action='store', default=5, type=int,
                    help="Visualize every n-th timestep.")

args = parser.parse_args()
ser_ref_levels = args.refine_serial
par_ref_levels = args.refine_parallel
order = args.order
dt = args.time_step
t_final = args.t_final
alpha = args.alpha
kappa = args.kappa
visualization = args.visualization
vis_steps = args.visualization_steps
ode_solver_type = args.ode_solver
if myid == 0:
    parser.print_options(args)

device = mfem.Device('cpu')
if myid == 0:
    device.Print()

# 3. Read the serial mesh from the given mesh file on all processors. We can
#    handle triangular, quadrilateral, tetrahedral and hexahedral meshes
#    with the same code.
meshfile = expanduser(join(dirname(__file__), '..', 'data', args.mesh))
mesh = mfem.Mesh(meshfile, 1, 1)
dim = mesh.Dimension()

# 4. Define the ODE solver used for time integration. Several implicit
#    singly diagonal implicit Runge-Kutta (SDIRK) methods, as well as
#    explicit Runge-Kutta methods are available.
if ode_solver_type == 1:
    ode_solver = BackwardEulerSolver()
elif ode_solver_type == 2:
    ode_solver = mfem.SDIRK23Solver(2)
elif ode_solver_type == 3:
    ode_solver = mfem.SDIRK33Solver()
elif ode_solver_type == 11:
    ode_solver = ForwardEulerSolver()
elif ode_solver_type == 12:
    ode_solver = mfem.RK2Solver(0.5)
elif ode_solver_type == 13:
    ode_solver = mfem.RK3SSPSolver()
elif ode_solver_type == 14:
    ode_solver = mfem.RK4Solver()
elif ode_solver_type == 22:
    ode_solver = mfem.ImplicitMidpointSolver()
elif ode_solver_type == 23:
    ode_solver = mfem.SDIRK23Solver()
elif ode_solver_type == 24:
    ode_solver = mfem.SDIRK34Solver()
else:
    print("Unknown ODE solver type: " + str(ode_solver_type))
    exit
# 5. Refine the mesh in serial to increase the resolution. In this example
#    we do 'ser_ref_levels' of uniform refinement, where 'ser_ref_levels' is
#    a command-line parameter.
for lev in range(ser_ref_levels):
    mesh.UniformRefinement()

# 6. Define a parallel mesh by a partitioning of the serial mesh. Refine
#    this mesh further in parallel to increase the resolution. Once the
#    parallel mesh is defined, the serial mesh can be deleted.
pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
del mesh
for x in range(par_ref_levels):
    pmesh.UniformRefinement()

# 7. Define the vector finite element space representing the current and the
#    initial temperature, u_ref.
fe_coll = mfem.H1_FECollection(order, dim)
fespace = mfem.ParFiniteElementSpace(pmesh, fe_coll)

fe_size = fespace.GlobalTrueVSize()
if myid == 0:
    print("Number of temperature unknowns: " + str(fe_size))
u_gf = mfem.ParGridFunction(fespace)

# 8. Set the initial conditions for u. All boundaries are considered
#    natural.
u_0 = InitialTemperature()
u_gf.ProjectCoefficient(u_0)
u = mfem.Vector()
u_gf.GetTrueDofs(u)

# 9. Initialize the conduction operator and the visualization.
oper = ConductionOperator(fespace, alpha, kappa, u)
u_gf.SetFromTrueDofs(u)

smyid = '{:0>6d}'.format(myid)
mesh_name = "ex16-mesh."+smyid
sol_name = "ex16-init."+smyid
pmesh.Print(mesh_name, 8)
u_gf.Save(sol_name, 8)

if visualization:
    sout = mfem.socketstream("localhost", 19916)
    sout.send_text("parallel " + str(num_procs) + " " + str(myid))
    sout.precision(8)
    sout.send_solution(pmesh, u_gf)
    sout.send_text("pause")
    sout.flush()
    if myid == 0:
        print("GLVis visualization paused.")
        print(" Press space (in the GLVis window) to resume it.")

# 8. Perform time-integration (looping over the time iterations, ti, with a
#    time-step dt).
ode_solver.Init(oper)
t = 0.0
ti = 1
last_step = False
while not last_step:

    if (t + dt >= t_final - dt/2):
        last_step = True

    t, dt = ode_solver.Step(u, t, dt)

    if (last_step or (ti % vis_steps) == 0):
        if myid == 0:
            print("step " + str(ti) + ", t = " + "{:g}".format(t))
        u_gf.SetFromTrueDofs(u)
        if (visualization):
            sout.send_text("parallel " + str(num_procs) + " " + str(myid))
            sout.send_solution(pmesh, u_gf)
            sout.flush()
    ti = ti + 1
    oper.SetParameters(u)

# 11. Save the final solution in parallel. This output can be viewed later
#     using GLVis: "glvis -np <np> -m ex16-mesh -g ex16-final".
sol_name = "ex16-final."+smyid
u_gf.Save(sol_name, 8)



def read_mat(file):
    fid = open(file, 'r')
    r = []
    c = []
    realdata = []
    imagdata = []

    for l in fid.readlines():
        xx = l.split()
        r.append(int(xx[0]))
        c.append(int(xx[1]))
        realdata.append(float(xx[2]))
        if len(xx) > 3:
           imagdata.append(float(xx[3]))

    return r, c, realdata, imagdata



# ----------------------------------------------------------------------------------------
# Routines for hypre
# ----------------------------------------------------------------------------------------

import sys
import os
import re
import subprocess

__all__ = ["cmake_make_hypre"]

from build_utils import *
from build_consts import *

import build_globals as bglb


def cmake_make_hypre():
    '''
    build hypre
    '''
    if bglb.verbose:
        print("Building hypre")

    cmbuild = 'cmbuild'
    path = os.path.join(extdir, 'hypre', 'src', cmbuild)
    if os.path.exists(path):
        print("working directory already exists!")
    else:
        os.makedirs(path)

    pwd = chdir(path)

    cmake_opts = {'DBUILD_SHARED_LIBS': '1',
                  'DHYPRE_INSTALL_PREFIX': bglb.hypre_prefix,
                  'DHYPRE_ENABLE_SHARED': '1',
                  'DCMAKE_C_FLAGS': '-fPIC',
                  'DCMAKE_INSTALL_PREFIX': bglb.hypre_prefix,
                  'DCMAKE_INSTALL_NAME_DIR': "@rpath", }
    if bglb.verbose:
        cmake_opts['DCMAKE_VERBOSE_MAKEFILE'] = '1'

    if bglb.enable_cuda and bglb.enable_cuda_hypre:
        # in this case, settitng CMAKE_C_COMPILER
        # causes "mpi.h" not found error. For now, letting CMAKE
        # to find MPI
        cmake_opts['DHYPRE_WITH_CUDA'] = '1'
        if bglb.cuda_arch != '':
            cmake_opts['DCMAKE_CUDA_ARCHITECTURES'] = bglb.cuda_arch
    else:
        cmake_opts['DCMAKE_C_COMPILER'] = bglb.mpicc_command

    cmake('..', **cmake_opts)

    make('hypre')
    make_install('hypre')

    os.chdir(pwd)



'''
   MFEM example 13p

    This example code solves the Maxwell (electromagnetic)
    eigenvalue problem curl curl E = lambda E with homogeneous
    Dirichlet boundary conditions E x n = 0.

   See c++ version in the MFEM library for more detail 
'''
from numpy import sin, cos, exp
import numpy as np
import sys
import mfem.par as mfem
from mfem.par import intArray, doubleArray
from os.path import expanduser, join, dirname

from mpi4py import MPI
num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank


ser_ref_levels = 2
par_ref_levels = 1
order = 1
nev = 5
visualization = 1

device = mfem.Device('cpu')
if myid == 0:
    device.Print()

num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank

meshfile = expanduser(join(dirname(__file__), '..', 'data', 'beam-tet.mesh'))
mesh = mfem.Mesh(meshfile, 1, 1)

dim = mesh.Dimension()

for k in range(ser_ref_levels):
    mesh.UniformRefinement()

pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
for k in range(par_ref_levels):
    pmesh.UniformRefinement()

fec = mfem.ND_FECollection(order, dim)
fespace = mfem.ParFiniteElementSpace(pmesh, fec)
size = fespace.GlobalTrueVSize()
if (myid == 0):
    print("Number of unknowns: " + str(size))

one = mfem.ConstantCoefficient(1.0)
ess_bdr = intArray()
if (pmesh.bdr_attributes.Size() > 0):
    ess_bdr.SetSize(pmesh.bdr_attributes.Max())
    ess_bdr.Assign(1)

a = mfem.ParBilinearForm(fespace)
a.AddDomainIntegrator(mfem.CurlCurlIntegrator(one))

if (pmesh.bdr_attributes.Size() == 0):
    # Add a mass term if the mesh has no boundary, e.g. periodic mesh or
    # closed surface.
    a.AddDomainIntegrator(mfem.VectorFEMassIntegrator(one))

a.Assemble()
a.EliminateEssentialBCDiag(ess_bdr, 1.0)
a.Finalize()

m = mfem.ParBilinearForm(fespace)
m.AddDomainIntegrator(mfem.VectorFEMassIntegrator(one))
m.Assemble()
# shift the eigenvalue corresponding to eliminated dofs to a large value
m.EliminateEssentialBCDiag(ess_bdr, 2.3e-308)
m.Finalize()

A = a.ParallelAssemble()
M = m.ParallelAssemble()

ams = mfem.HypreAMS(A, fespace)
ams.SetPrintLevel(0)
ams.SetSingularProblem()

ame = mfem.HypreAME(MPI.COMM_WORLD)
ame.SetNumModes(nev)
ame.SetPreconditioner(ams)
ame.SetMaxIter(100)
ame.SetTol(1e-8)
ame.SetPrintLevel(1)
ame.SetMassMatrix(M)
ame.SetOperator(A)

eigenvalues = doubleArray()
ame.Solve()
ame.GetEigenvalues(eigenvalues)
x = mfem.ParGridFunction(fespace)


smyid = '{:0>6d}'.format(myid)
mesh_name = "mesh."+smyid


pmesh.Print(mesh_name, 8)

for i in range(nev):
    sol_name = "mode_"+str(i).zfill(2)+"."+smyid
    x.Assign(ame.GetEigenvector(i))
    x.Save(sol_name, 8)


if visualization:
    for i in range(nev):
        mode_sock = mfem.socketstream("localhost", 19916)
        mode_sock.precision(8)

        if (myid == 0):
            print("Eigenmode " + str(i+1) + '/' + str(nev) +
                  ", Lambda = " + str(eigenvalues[i]))

        x.Assign(ame.GetEigenvector(i))
        mode_sock << "parallel " << num_procs << " " << myid << "\n"
        mode_sock << "solution\n" << pmesh << x
        mode_sock.flush()
        mode_sock << "window_title 'Eigenmode " << i+1 << '/' << nev
        mode_sock << ", Lambda = " << "{:g}".format(eigenvalues[i]) << "'"
        mode_sock.endline()

        if (myid == 0):
            from builtins import input
            c = input("press (q)uit or (c)ontinue --> ")
        else:
            c = None
        c = MPI.COMM_WORLD.bcast(c, root=0)
        if (c != 'c'):
            sys.exit()



from __future__ import print_function
import os
import sys

def run_test():
    print("Test mesh module")
    Nvert = 6; Nelem = 8; Nbelem=2
    
    mesh = mfem.Mesh(2, Nvert, Nelem, 2, 3)
    tri_v = [[1.,  0.,  0.], [0.,  1.,  0.], [-1.,  0.,  0.],
             [0., -1.,  0.], [0.,  0.,  1.], [ 0.,  0., -1.]]
    tri_e = [[0, 1, 4], [1, 2, 4], [2, 3, 4], [3, 0, 4],
             [1, 0, 5], [2, 1, 5], [3, 2, 5], [0, 3, 5]]
    tri_l = [[1,4], [1,2]]
    
    for j in range(Nvert):
        mesh.AddVertex(tri_v[j])
    for j in range(Nelem):
        mesh.AddTriangle(tri_e[j], 1)
    for j in range(Nbelem):
        mesh.AddBdrSegment(tri_l[j], 1)
        
    mesh.FinalizeTriMesh(1,1, True)
    t = mesh.GetEdgeVertexTable()
    t.Print()
    
if __name__=='__main__':
    if len(sys.argv) > 1 and sys.argv[1] == '-p':   
        import mfem.par as mfem
    else:
        import mfem.ser as mfem
        
    run_test()



import os

txt1 = "//\n"
txt2 = "// Copyright (c) 2020-2025, Princeton Plasma Physics Laboratory, All rights reserved.\n"

def find_i_file(path):

    for x in os.listdir(path):
        if not x.endswith(".i"):
            continue
               
        fname = os.path.join(path, x)
        print("processing: ", x)
        fid = open(fname)
        lines = fid.readlines()
        fid.close()
        
        if lines[1].startswith("// Copyright"):
            lines[1] = txt2
        else:
            lines = [txt1, txt2, txt1]+lines


        fid = open(fname, "w")
        fid.write("".join(lines))
        fid.close()
        
    

if __name__ == '__main__':
    find_i_file("../mfem/_ser")
    find_i_file("../mfem/_par")
    find_i_file("../mfem/common")        
    



'''
   MFEM example 18
      This is a version of Example 18 with a simple adaptive mesh
      refinement loop. 
      See c++ version in the MFEM library for more detail 

   Sample runs:

       python ex18.py -p 1 -r 2 -o 1 -s 3
       python ex18.py -p 1 -r 1 -o 3 -s 4
       python ex18.py -p 1 -r 0 -o 5 -s 6
       python ex18.py -p 2 -r 1 -o 1 -s 3 -mf
       python ex18.py -p 2 -r 0 -o 3 -s 3 -mf
'''
from mfem.common.arg_parser import ArgParser
import mfem.ser as mfem
from mfem.ser import intArray
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sqrt, pi, cos, sin, hypot, arctan2
from scipy.special import erfc

from ex18_common import (EulerMesh,
                         EulerInitialCondition,
                         DGHyperbolicConservationLaws)


def run(problem=1,
        ref_levels=1,
        order=3,
        ode_solver_type=4,
        t_final=2.0,
        dt=-0.01,
        cfl=0.3,
        visualization=True,
        vis_steps=50,
        preassembleWeakDiv=False,
        meshfile=''):

    specific_heat_ratio = 1.4
    gas_constant = 1.0
    IntOrderOffset = 1

    # 2. Read the mesh from the given mesh file. This example requires a 2D
    #    periodic mesh, such as ../data/periodic-square.mesh.

    mesh = EulerMesh(meshfile, problem)
    dim = mesh.Dimension()
    num_equation = dim + 2

    # Refine the mesh to increase the resolution. In this example we do
    # 'ref_levels' of uniform refinement, where 'ref_levels' is a
    # command-line parameter.
    for lev in range(ref_levels):
        mesh.UniformRefinement()

    # 3. Define the ODE solver used for time integration. Several explicit
    #    Runge-Kutta methods are available.
    ode_solver = None
    if ode_solver_type == 1:
        ode_solver = mfem.ForwardEulerSolver()
    elif ode_solver_type == 2:
        ode_solver = mfem.RK2Solver(1.0)
    elif ode_solver_type == 3:
        ode_solver = mfem.RK3SSPSolver()
    elif ode_solver_type == 4:
        ode_solver = mfem.RK4Solver()
    elif ode_solver_type == 6:
        ode_solver = mfem.RK6Solver()
    else:
        print("Unknown ODE solver type: " + str(ode_solver_type))
        exit

    # 4. Define the discontinuous DG finite element space of the given
    #    polynomial order on the refined mesh.

    fec = mfem.DG_FECollection(order, dim)
    # Finite element space for a scalar (thermodynamic quantity)
    fes = mfem.FiniteElementSpace(mesh, fec)
    # Finite element space for a mesh-dim vector quantity (momentum)
    dfes = mfem.FiniteElementSpace(mesh, fec, dim, mfem.Ordering.byNODES)
    # Finite element space for all variables together (total thermodynamic state)
    vfes = mfem.FiniteElementSpace(
        mesh, fec, num_equation, mfem.Ordering.byNODES)

    assert fes.GetOrdering() == mfem.Ordering.byNODES, "Ordering must be byNODES"
    print("Number of unknowns: " + str(vfes.GetVSize()))

    # 5. Define the initial conditions, save the corresponding mesh and grid
    #    functions to files. These can be opened with GLVis using:
    #    "glvis -m euler-mesh.mesh -g euler-1-init.gf" (for x-momentum).
    u0 = EulerInitialCondition(problem,
                               specific_heat_ratio,
                               gas_constant)
    sol = mfem.GridFunction(vfes)
    sol.ProjectCoefficient(u0)

    # (Python note): GridFunction pointing to the subset of vector FES.
    #  sol is Vector with dim*fes.GetNDofs()
    #  Since sol.GetDataArray() returns numpy array pointing to the data, we make
    #  Vector from a sub-vector of the returned numpy array and pass it to GridFunction
    #  constructor.

    mom = mfem.GridFunction(dfes, mfem.Vector(
        sol.GetDataArray()[fes.GetNDofs():]))
    mesh.Print("euler-mesh.mesh", 8)

    for k in range(num_equation):
        uk = mfem.GridFunction(fes, mfem.Vector(
            sol.GetDataArray()[k*fes.GetNDofs():]))
        sol_name = "euler-" + str(k) + "-init.gf"
        uk.Save(sol_name, 8)

    # 6. Set up the nonlinear form with euler flux and numerical flux
    flux = mfem.EulerFlux(dim, specific_heat_ratio)
    numericalFlux = mfem.RusanovFlux(flux)
    formIntegrator = mfem.HyperbolicFormIntegrator(
        numericalFlux, IntOrderOffset)

    euler = DGHyperbolicConservationLaws(vfes, formIntegrator,
                                         preassembleWeakDivergence=preassembleWeakDiv)

    # 7. Visualize momentum with its magnitude
    if (visualization):
        sout = mfem.socketstream("localhost", 19916)
        sout.precision(8)
        sout << "solution\n" << mesh << mom
        sout << "window_title 'momentum, t = 0'\n"
        sout << "view 0 0\n"  # view from top
        sout << "keys jlm\n"  # turn off perspective and light, show mesh
        sout << "pause\n"
        sout.flush()
        print("GLVis visualization paused.")
        print(" Press space (in the GLVis window) to resume it.")

    # 8. Time integration
    hmin = np.inf
    if (cfl > 0):
        hmin = min([mesh.GetElementSize(i, 1) for i in range(mesh.GetNE())])

        # Find a safe dt, using a temporary vector. Calling Mult() computes the
        # maximum char speed at all quadrature points on all faces (and all
        # elements with -mf).
        z = mfem.Vector(sol.Size())
        euler.Mult(sol, z)

        max_char_speed = euler.GetMaxCharSpeed()
        dt = cfl * hmin / max_char_speed / (2 * order + 1)

    t = 0.0
    euler.SetTime(t)
    ode_solver.Init(euler)

    # Integrate in time.
    done = False
    ti = 0
    while not done:
        dt_real = min(dt, t_final - t)

        t, dt_real = ode_solver.Step(sol, t, dt_real)

        if (cfl > 0):
            max_char_speed = euler.GetMaxCharSpeed()
            dt = cfl * hmin / max_char_speed / (2*order+1)
        ti = ti+1
        done = (t >= t_final - 1e-8*dt)
        if (done or ti % vis_steps == 0):
            print("time step: " + str(ti) + ", time: " + "{:g}".format(t))
            if (visualization):
                sout << "window_title 'momentum, t = " << "{:g}".format(
                    t) << "'\n"
                sout << "solution\n" << mesh << mom
                sout.flush()

    #  8. Save the final solution. This output can be viewed later using GLVis:
    #    "glvis -m euler.mesh -g euler-1-final.gf".
    mesh.Print("euler-mesh-final.mesh", 8)
    for k in range(num_equation):
        uk = mfem.GridFunction(fes, mfem.Vector(
            sol.GetDataArray()[k*fes.GetNDofs():]))
        sol_name = "euler-" + str(k) + "-final.gf"
        uk.Save(sol_name, 8)

    print(" done")
    # 9. Compute the L2 solution error summed for all components.
    if (t_final == 2.0):
        error = sol.ComputeLpError(2., u0)
        print("Solution error: " + "{:g}".format(error))


if __name__ == "__main__":

    parser = ArgParser(description='Ex18')
    parser.add_argument('-m', '--mesh',
                        default='',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-p', '--problem',
                        action='store', default=1, type=int,
                        help='Problem setup to use. See options in velocity_function().')
    parser.add_argument('-r', '--refine',
                        action='store', default=1, type=int,
                        help="Number of times to refine the mesh uniformly.")
    parser.add_argument('-o', '--order',
                        action='store', default=3, type=int,
                        help="Finite element order (polynomial degree)")
    parser.add_argument('-s', '--ode_solver',
                        action='store', default=4, type=int,
                        help="ODE solver: 1 - Forward Euler,\n\t" +
                        "            2 - RK2 SSP, 3 - RK3 SSP, 4 - RK4, 6 - RK6.")
    parser.add_argument('-tf', '--t_final',
                        action='store', default=2.0, type=float,
                        help="Final time; start time is 0.")
    parser.add_argument("-dt", "--time_step",
                        action='store', default=-0.01, type=float,
                        help="Time step.")
    parser.add_argument('-c', '--cfl_number',
                        action='store', default=0.3, type=float,
                        help="CFL number for timestep calculation.")
    parser.add_argument('-novis', '--no_visualization',
                        action='store_true', default=False,
                        help='Disable GLVis visualization')
    parser.add_argument("-ea", "--element-assembly-divergence",
                        action='store_true', default=False,
                        help="Weak divergence assembly level\n" +
                        "    ea - Element assembly with interpolated")
    parser.add_argument("-mf", "--matrix-free-divergence",
                        action='store_true', default=False,
                        help="Weak divergence assembly level\n" +
                        "    mf - Nonlinear assembly in matrix-free manner")
    parser.add_argument('-vs', '--visualization-steps',
                        action='store', default=50, type=float,
                        help="Visualize every n-th timestep.")

    args = parser.parse_args()

    visualization = not args.no_visualization

    if (not args.matrix_free_divergence and
            not args.element_assembly_divergence):
        args.element_assembly_divergence = True
        args.matrix_free_divergence = False
        preassembleWeakDiv = True

    elif args.element_assembly_divergence:
        args.matrix_free_divergence = False
        preassembleWeakDiv = True

    elif args.matrix_free_divergence:
        args.element_assembly_divergence = False
        preassembleWeakDiv = False

    parser.print_options(args)

    run(problem=args.problem,
        ref_levels=args.refine,
        order=args.order,
        ode_solver_type=args.ode_solver,
        t_final=args.t_final,
        dt=args.time_step,
        cfl=args.cfl_number,
        visualization=visualization,
        vis_steps=args.visualization_steps,
        preassembleWeakDiv=preassembleWeakDiv,
        meshfile=args.mesh)
